from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from builtins import open
from builtins import next
from builtins import map
from builtins import int
from future import standard_library
standard_library.install_aliases()
from builtins import object
import ctypes
import datetime
import time
import re
import codecs

from binascii import hexlify
from hashlib import sha1
from os import makedirs, stat, getpid, listdir, unlink, kill, name, rename, rmdir, strerror
from os.path import dirname, exists, isdir, expanduser, abspath, join, realpath, getmtime
from shutil import move, copyfile
from subprocess import Popen, check_output, STDOUT
from sys import version_info

if version_info[0] >= 3:
    from os import replace

from requests import __version__ as requests_version, Session
from requests.adapters import HTTPAdapter


"""
Assorted utilities.
"""


def create_parents(path):
    """
    Make sure that the directories in the path exist.
    """
    dir = dirname(path)
    if not exists(dir):
        makedirs(dir)
    if not isdir(dir):
        raise Exception('"%s" is not a directory' % dir)


def touch(path):
    """
    Create or 'modify' a file.
    """
    create_parents(path)
    open(path, 'a').close()


def safe_unlink(path):
    """
    Delete a file if it exists.
    """
    if path is not None and exists(path):
        try:
            unlink(path)
        except PermissionError:
            pass  # file still in use on windows


def check_cmd(config, param, name):
    """
    Check the command exists and, if not, inform the user.
    """
    from .args import FORCECMD
    value = config.arg(param)
    if windows() and '/' in value:
        config.log.warn('Replacing slashes with back-slashes in "%s"' % value)
        value = re.sub(r'/', r'\\', value)
    if not config.arg(FORCECMD):
        cmd = '%s -h' % value
        try:
            check_output(cmd, stderr=STDOUT, shell=True)
            return value
        except Exception as e:
            config.log.error('Command "%s" failed' % cmd)
            config.log.error('Install %s or configure %s correctly' % (name, param))
            raise Exception('Cannot find %s (using %s)' % (name, cmd))
    else:
        config.log.warn('Not checking command %s' % name)
        return value


def canonify(path):
    """
    Expand the path so it's repeatable.
    """
    return realpath(abspath(expanduser(path)))


def canonify_dir_and_make(path):
    """
    We need to canonify and make sure some dirs exist.
    """
    path = canonify(path)
    if not exists(path):
        makedirs(path)
    if not isdir(path):
        raise Exception('%s is not a directory')
    return path


def run(cmd, log, uncouple=False):
    """
    We can't use subprocess.run() because it doesn't exist for 2.7.
    """
    log.debug('Running "%s"' % cmd)
    if uncouple:
        if version_info[0] >= 3:
            Popen(cmd, shell=True, close_fds=True, start_new_session=True)
        else:
            Popen(cmd, shell=True, close_fds=True)
    else:
        process = Popen(cmd, shell=True)
        process.wait()
        if process.returncode:
            raise Exception('Command "%s" failed' % cmd)


def check_leap(enabled, expire, file, url, timeout, retries, log):
    """
    Download a file if none exists or it is more than 3 months old.

    Returns the file name or 'NONE' - the value to be passed to the mseedindex command.
    """
    if enabled:
        file = canonify(file)
        if exists(file):
            statinfo = stat(file)
            age = int(time.time()) - statinfo.st_atime
            log.debug('%s is %ds old' % (file, age))
            download = age > expire * 24 * 60 * 60
        else:
            download= True

        if download:
            get_to_file(url, file, timeout, retries, log, unique=False)

        if valid_leapfile(file, log):
            return file
        else:
            log.warning('Invalid leap second file will be ignored: %s' % file)
            return 'NONE'
    else:
        return 'NONE'

def valid_leapfile(file, log):
    """
    Perform simple validation of an IETF leap second file.

    Validation: all non-comment or blank lines should contain
    1) an epoch time stamp, followed by
    2) a offset in seconds
    Both fields are all digits separated by whitespace with optional
    trailing comment, for example:
    '3692217600      37      # 1 Jan 2017'

    Returns the file name on success or None if an unrecognized line is found
    """
    twodigits = re.compile("^\d+\s+\d+$")

    with open(file) as fp:
        for line in fp.readlines():
            trimmed = line.partition('#')[0].rstrip('\n\r').rstrip()

            if len(trimmed) == 0:
                continue

            if not twodigits.match(trimmed):
                log.warning('Unrecognized line in leap second file: "%s"' % line.rstrip('\n\r'))
                return None

    return file

def hash(text):
    """
    SHA1 hash as hex.
    """
    hash = sha1()
    hash.update(text.encode('utf-8'))
    return hexlify(hash.digest()).decode('ascii')

def uniqueish(prefix, salt, pid=None):
    """
    Generate a unique(ish) name, from a prefix, text (hashed) and pid.
    """
    if pid is None:
        pid = getpid()
    return '%s_%s_%d' % (prefix, hash(salt)[:6], pid)


def unique_filename(path):
    """
    Append a count until we find a unique name.
    """
    if exists(path):
        count = 0
        while True:
            count += 1
            new_path = '%s.%d' % (path, count)
            if not exists(new_path):
                return new_path
    else:
        return path


def unique_path(dir, filename, salt):
    """
    Generate a unique path to a temporary file.
    """
    name = uniqueish(filename, salt)
    return unique_filename(join(dir, name))

def _stream_output(request, down, unique=True):
    # special case empty return.  this avoids handling empty files elsewhere
    # which isn't a 'serious' problem, but causes ugly logging
    if request.status_code == 204:
        return None, lambda: None
    else:
        down = canonify(down)
        create_parents(down)
        if unique:
            down = unique_filename(down)
        with open(down, 'wb') as output:
            for chunk in request.iter_content(chunk_size=1024):
                if chunk:
                    output.write(chunk)
        return down, request.raise_for_status


def _session(retries):
    """
    Ugliness required by requests lib to set max retries.
    (We don't really care about efficiency to the point where we need to re-use the session)
    """
    # https://stackoverflow.com/questions/21371809/cleanly-setting-max-retries-on-python-requests-get-or-post-method
    session = Session()
    http_adapter = HTTPAdapter(max_retries=retries)
    https_adapter = HTTPAdapter(max_retries=retries)
    session.mount('http://', http_adapter)
    session.mount('https://', https_adapter)

    # Create a User-Agent header with package, requests and Python identifiers
    from rover import __version__
    user_agent = 'rover/%s python-requests/%s Python/%s' % \
                 (__version__, requests_version, ".".join(map(str, version_info[:3])))
    session.headers.update({'User-Agent': user_agent})

    return session


def get_to_file(url, down, timeout, retries, log, unique=True):
    """
    Execute an HTTP GET request, with output to a file.

    Returns (path, lambda)
    where path is the path to the file (possibly None if no data downloaded)
          lambda() (ie when called) will raise an exception on HTTP error
    this gives the caller both the results (which may contain error msg)
    and the error exception.
    """
    log.info('Downloading %s from %s' % (down, url))
    request = _session(retries).get(url, stream=True, timeout=timeout)
    return _stream_output(request, down, unique=unique)


def post_to_file(url, up, down, timeout, retries, log, unique=True):
    """
    Execute an HTTP POST request, with output to a file.

    Returns (path, lambda)
    where path is the path to the file (possibly None if no data downloaded)
          lambda() (ie when called) will raise an exception on HTTP error
    this gives the caller both the results (which may contain error msg)
    and the error exception.
    """
    up = canonify(up)
    log.info('Downloading %s from %s with %s' % (down, url, up))
    with open(up, 'rb') as input:
        request = _session(retries).post(url, stream=True, data=input, timeout=timeout)
    return _stream_output(request, down, unique=unique)


def clean_old_files(dir, age_secs, match, log):
    """
    Delete old files that match the predicate.
    """
    if exists(dir):
        for file in listdir(dir):
            if match(file):
                try:
                    if time.time() - getmtime(file) > age_secs:
                        log.warn('Deleting old %s' % file)
                except:   # py2.7 no FileNotFound
                    pass  # was deleted from under us


def match_prefixes(*prefixes):
    """
    Match predicate (see above) using a prefix.
    """
    def match(name):
        for prefix in prefixes:
            if name.startswith(prefix):
                return True
        return False
    return match


class PushBackIterator(object):
    """
    Modify an iterator so that a (single) value can be pushed back
    and will be returned next iteration.
    """

    def __init__(self, iter):
        self._iter = iter
        self._pushed = None

    def push(self, value):
        if self._pushed:
            raise Exception('Cannot push multiple values')
        self._pushed = value

    def __iter__(self):
        return self

    def __next__(self):
        if self._pushed:
            value, self._pushed = self._pushed, None
        else:
            value = next(self._iter)
        return value


ZERO = datetime.timedelta(0)


class UTC(datetime.tzinfo):
    """UTC timezone (needed for Py2.7)"""

    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return ZERO


utc = UTC()
EPOCH = datetime.datetime.utcfromtimestamp(0)
EPOCH_UTC = EPOCH.replace(tzinfo=utc)


def assert_valid_time(log, time):
    """
    Check timestamp format.
    """
    if re.match(r'^\d{4}-\d{2}-\d{2}(T\d{2}:\d{2}:\d{2}(\.\d+)?)?$', time):
        return time
    else:
        msg = 'Invalid time format "%s"' % time
        log.error(msg)
        raise Exception(msg)


def format_epoch(epoch):
    """
    Format an epoch in the standard format.
    """
    dt = datetime.datetime.fromtimestamp(epoch, utc)
    return datetime.datetime.strftime(dt, '%Y-%m-%dT%H:%M:%S.%f')


def format_day_epoch(epoch):
    """
    Format an epoch as a date, without time.
    """
    dt = datetime.datetime.fromtimestamp(epoch, utc)
    return datetime.datetime.strftime(dt, '%Y-%m-%d')


def format_year_day_epoch(epoch):
    """
    Format an epoch as year and day
    """
    dt = datetime.datetime.fromtimestamp(epoch, utc)
    return datetime.datetime.strftime(dt, '%Y-%j')


def format_time_epoch(epoch):
    """
    Format an epoch, with time to seconds
    """
    dt = datetime.datetime.fromtimestamp(epoch, utc)
    return datetime.datetime.strftime(dt, '%Y-%m-%dT%H:%M:%S')


def format_time_epoch_local(epoch):
    """
    Format an epoch, with time to seconds
    """
    dt = datetime.datetime.fromtimestamp(epoch)
    return datetime.datetime.strftime(dt, '%Y-%m-%dT%H:%M:%S')


def parse_epoch(date):
    """
    Parse a date in the standard formats
    """
    if date.endswith('Z'):
        date = date[:-1]
    try:
        dt = datetime.datetime.strptime(date, '%Y-%m-%dT%H:%M:%S.%f')
    except ValueError:
        try:
            dt = datetime.datetime.strptime(date, '%Y-%m-%dT%H:%M:%S')
        except ValueError:
            try:
                dt = datetime.datetime.strptime(date, '%Y-%m-%dT%H:%M')
            except ValueError:
                dt = datetime.datetime.strptime(date, '%Y-%m-%d')
    return (dt - EPOCH).total_seconds()


def tidy_timestamp(log, timestamp):
    try:
        tidied = format_epoch(parse_epoch(timestamp))
        if tidied != timestamp:
            log.debug('Tidied timestamp: "%s" -> "%s"' % (timestamp, tidied))
        return tidied
    except:
        msg = 'Cannot parse timestamp "%s"' % timestamp
        log.error(msg)
        raise Exception(msg)


def in_memory(iterator):
    """
    Pull an entire iterator into memory and then re-expose as an iterator.
    """
    return iter(list(iterator))


STATION = 'station'
NETWORK = 'network'
CHANNEL = 'channel'
LOCATION = 'location'


def build_file(log, path, args):
    """
    Given a N_S_L_C or net=... and start/end times, construct an input file in
    the correct (availability service) format.
    """
    # just go crazy because any error is caught by the caller and changed into a 'bad syntax' error
    if '_' in args[0]:
        (n, s, l, c) = (code if code else '--' for code in args[0].split('_'))
        args = ['net=' + n, 'sta=' + s, 'loc=' + l, 'cha=' + c] + args[1:]
    sncl = {NETWORK: '*', STATION: '*', LOCATION: '*', CHANNEL: '*'}
    count = 0
    while args and '=' in args[0]:
        arg = args.pop(0)
        name, value = arg.split('=')
        for key in sncl.keys():
            if key.startswith(name):
                sncl[key] = value
                count += 1
    assert count
    assert len(args) < 3
    parts = [sncl[NETWORK], sncl[STATION], sncl[LOCATION], sncl[CHANNEL]]
    parts += args
    with open(path, 'w') as req:
        print(' '.join(parts), file=req)


def sort_file_inplace(log, path, temp_dir, sort_in_python):
    """
    Sort file using UNIX command line utility.
    """
    done = False
    try:
        if not sort_in_python:
            _os_sort(log, path, temp_dir)
            done = True
    except Exception as e:
        log.warn('OS sorting failed (%s) using python fallback' % e)
    if not done:
        _python_sort(log, path)


def _python_sort(log, path):
    log.debug('Sorting %s in memory' % path)
    with open(path, 'r') as source:
        lines = source.readlines()
    with open(path, 'w') as dest:
        for line in sorted(lines):
            print(line.rstrip(), file=dest)


def _os_sort(log, path, temp_dir):
    sorted_path = unique_path(temp_dir, 'rover_sort', path)
    log.debug('Sorting %s into %s' % (path, sorted_path))
    run('sort %s > %s' % (path, sorted_path), log)
    safe_unlink(path)
    move(sorted_path, path)


def process_exists(pid):
    """
    Check whether the given PID exists.
    """
    if windows():
        # https://stackoverflow.com/questions/17620833/check-if-pid-exists-on-windows-with-python-without-requiring-libraries
        kernel32 = ctypes.windll.kernel32
        process = kernel32.OpenProcess(0x100000, 0, pid)
        if process:
            kernel32.CloseHandle(process)
            return True
        else:
            return False
    else:
        try:
            kill(pid, 0)
            return True
        except OSError:
            return False


def windows():
    """
    Are we running on windows?
    """
    return name in ('Windows', 'nt')


def diagnose_error(log, error, request, response, copied=True):
    # avoid import loop
    from .args import mm, VERBOSITY, NO, DELETEFILES
    log.error(error)
    log.error('Response contents (max 10 lines) are listed below:')
    log_file_contents(response, log, 10)
    log.error('Please pay special attention to the first lines of the message - ' +
              'they often contains useful information.')
    log.error('Request contents (max 10 lines) are listed below:')
    log_file_contents(request, log, 10)
    log.error('The request is either provided by the user or created from the user input.')
    if copied:
        log.error('To ensure consistency ROVER copies files.  ' +
                  'To see the paths and avoid deleting temporary copies re-run the command ' +
                  'with the %s 5 and %s%s options' % (mm(VERBOSITY), NO, DELETEFILES))


def log_file_contents(path, log, max_lines=10):
    log.info('Displaying contents of file %s:' % path)
    count = 0
    try:
        with codecs.open(path, encoding='utf-8', errors='strict') as input:
            for line in input:
                line = line.strip()
                if line:
                    log.error('> %s' % line)
                    count += 1
                    if count >= max_lines:
                        break
    except UnicodeDecodeError:
        log.error('File contents are not printable.')
    except IOError as e:
        log.error('Error opening file:', strerror(e.errno))


def calc_bytes(sizestring):
    """
    Calculate a size in bytes for the specified size string.  If the
    string is terminated with the following suffixes the specified
    scaling will be applied:

    'K' or 'k' : kilobytes - value * 1024
    'M' or 'm' : megabytes - value * 1024*1024
    'G' or 'g' : gigabytes - value * 1024*1024*1024

    Returns a size in bytes.
    """

    if sizestring.endswith('k') or sizestring.endswith('K'):
        return int(sizestring[:-1]) * 1024

    elif sizestring.endswith('m') or sizestring.endswith('M'):
        return int(sizestring[:-1]) * 1024 * 1024

    elif sizestring.endswith('g') or sizestring.endswith('G'):
        return int(sizestring[:-1]) * 1024 * 1024 * 1024

    else:
        return int(sizestring)


def null_fixer(log, line):
    return line


def iris_fixer(log, line):
    """
    Tidy and validate a request file line as used by the retrieve command.

    Empty lines and lines beginning with '#' are considered acceptable, but
    should not be submitted to a web service.

    Return the tidied request line on success.

    Return None on acceptable line that should not be included in a service request.

    Raise exception for unacceptable lines.
    """
    line = line.strip()

    if len(line) == 0:
        return None

    if line.startswith('#'):
        return None

    fields = line.split()

    if len(fields) != 6:
        raise Exception ("Unrecognized request line, not enough fields: '%s'" % line)

    # Acceptable source identifier codes contain only these characters
    acceptable_code = "[-_,A-Za-z0-9*?]"

    if not re.match(acceptable_code, fields[0]):
        raise Exception ("Unrecognized request line, invalid network code: '%s'" % line)

    if not re.match(acceptable_code, fields[1]):
        raise Exception ("Unrecognized request line, invalid station code: '%s'" % line)

    if not re.match(acceptable_code, fields[2]):
        raise Exception ("Unrecognized request line, invalid location code: '%s'" % line)

    if not re.match(acceptable_code, fields[3]):
        raise Exception ("Unrecognized request line, invalid channel code: '%s'" % line)

    # Tidy time values, allowing '*' as an exception (meaning "open" time)
    if fields[4] != '*':
        try:
            fields[4] = tidy_timestamp(log, fields[4])
        except:
            raise Exception ("Unrecognized request line, invalid start time: '%s'" % line)

    if fields[5] != '*':
        try:
            fields[5] = tidy_timestamp(log, fields[5])
        except:
            raise Exception ("Unrecognized request line, invalid end time: '%s'" % line)

    return " ".join(fields)


def fix_file_inplace(log, path, temp_dir, fixer=iris_fixer):
    temp_path = unique_path(temp_dir, 'rover_fixed_request', path)
    log.debug('Fixing %s in %s' % (path, temp_path))
    try:
        with open(temp_path, 'w') as output:
            with open(path, 'r') as input:
                for line in input.readlines():
                    line = line.rstrip()  # remove linefeed
                    line = fixer(log, line)
                    if line:
                        print(line, file=output)
        unlink(path)
        log.debug('Replacing %s with %s' % (path, temp_path))
        copyfile(temp_path, path)
    finally:
        safe_unlink(temp_path)


def atomic_move(log, src, dest):
    '''
    This uses an atomic primitive in 3.3+ and on unix.
    On Windows the best we can do is two operations, repeating if we catch an interrupt.
    https://bugs.python.org/issue8828
    '''
    if version_info[0] >= 3:
        log.debug('Moving %s to %s (atomic 3)' % (src, dest))
        replace(src, dest)
    else:
        if windows():
            log.debug('Moving %s to %s (windows)' % (src, dest))
            exception = None
            while exists(src):
                try:
                    safe_unlink(dest)
                    rename(src, dest)
                except KeyboardInterrupt as e:
                    log.debug('Caught interrupt; will re-throw once move complete')
                    exception = e
            if exception:
                raise exception
        else:
            log.debug('Moving %s to %s (atomic 2.7)' % (src, dest))
            rename(src, dest)


def remove_empty_folders(path, log):
    'Function to remove empty folders under root directory'
    if not isdir(path):
        return

    # remove empty subfolders
    files = listdir(path)
    if len(files):
        for f in files:
            fullpath = join(path, f)
            if isdir(fullpath):
                remove_empty_folders(fullpath, log)

    # if folder empty, delete it
    files = listdir(path)
    if len(files) == 0:
        log.debug("Removing empty folder: %s" % path)
        rmdir(path)

def dictionary_text_list(kwargs, prefix=""):
    cmds = []
    for command in kwargs.keys():
        cmds.append("  {0}{1:19}: {2}".format(prefix,
                                              command,
                                              kwargs[command][1]))
    return "\n".join(cmds)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import open
from builtins import range
from builtins import zip
from builtins import next
from builtins import map
from future import standard_library
standard_library.install_aliases()
from builtins import object
import operator
from functools import reduce
from re import sub, compile

from .args import mm, FORCEREQUEST
from .utils import parse_epoch

"""
Comparison of requests to check for overlap.
"""


def parse_request(path, regexp_only):
    """
    Read the request into memory, extracting the N_S_L_C and date data.
    """
    lines = []   # in-memory rather than a generator because we need to sort manually
    with open(path, 'r') as input:
        for line in input:
            try:
                line = line.strip()
                if line:
                    parts = line.split(' ')
                    sncl = parts[0:4]
                    if regexp_only:
                        if reduce(operator.or_, map(lambda x: '*' in x or '?' in x, sncl)):
                            # the outer sub here fixes things up so we match ? and * (we can't be explicit about
                            # * in the innermost sub because the middle sub would replace it).
                            sncl = list(map(lambda x: sub(r'\?', '?\\*',
                                                          sub(r'\*', '[A-Za-z0-9\-\?]*',
                                                              sub(r'\?', '[A-Za-z0-9\-\?]',
                                                                  x))), sncl))
                        else:
                            continue  # only regexps
                    dates = parts[4:]
                    assert len(dates) <= 2
                    while len(dates) < 2:
                        dates.append(None)
                    lines.append((' '.join(sncl), [tuple(dates)]))
            except:
                raise Exception(('Cannot parse "%s" in %s (experts can use %s ' +
                                 'at the risk of duplicating data in the repository)') %
                                (line, path, mm(FORCEREQUEST)))
    return lines


def unique(input):
    """
    If the same N_S_L_C occurs twice, merge the dates into a single entry.
    Needed for the fast comparison (of sorted files) to work correctly.
    """
    output = []
    for (sncl, dates) in sorted(input):
        if output and sncl == output[-1][0]:
            output[-1][1].append(dates[0])
        else:
            output.append((sncl, dates))
    return output


def overlapping_pair(pair1, pair2):
    """
    Check whether two pairs of dates (start and end) overlap.
    """
    # if either is open, they must overlap
    if not pair1 or not pair2:
        return True
    # if they're both lower bounds, they must overlap
    if len(pair1) == 1 and len(pair2) == 1:
        return True
    pair1 = list(map(parse_epoch, pair1))
    pair2 = list(map(parse_epoch, pair2))
    # dates1 is a range
    if len(pair1) == 2:
        # dates2 is a lower bound?
        if len(pair2) == 1:
            # if dates2 starts before dates1 ends, they overlap
            if pair2[0] < pair1[1]:
                return True
        else:
            # easier to find when they don't overlap, and then negate
            # they don't overlap if dates1 ends before dates2 starts, or vice versa
            if not (pair1[1] < pair2[0] or pair2[1] < pair1[0]):
                return True
    # dates1 is a lower bound
    else:
        # if dates1 starts before dates2 ends, they overlap
        if pair1[0] < pair2[1]:
            return True
    # if none of the above, no overlap
    return False


def overlapping_dates(dates1, dates2):
    """
    Check whether lists of pairs of dates contain any pairs that overlap.
    """
    # here dates are lists of tuples and we consider each possible pairing
    # which is O(n^2) but we don't expect to have many (i pray...)
    for pair1 in dates1:
        for pair2 in dates2:
            if overlapping_pair(pair1, pair2):
                return True
    return False


def format_dates(dates):
    """
    Format a list of date pairs, where the dates can be None if missing / open.
    """
    return ', '.join(map(lambda pair: ' - '.join(map(lambda date: date if date else 'open', pair)), dates))


class Regexps(object):
    """
    Encapsulate the regexps associated with a file.
    """

    def __init__(self, sncls, dates, path):
        self._sncls = sncls
        self._dates = dates
        self._path = path
        self._regexp = compile('(%s)' % '|'.join(sncls))

    def assert_no_overlap(self, sncl, dates, path):
        """
        Raise an exception on overlap.
        """
        # use a single "merged" regexp as gatekeeper and then run through each entry to find the
        # correct date.
        if self._regexp.match(sncl):
            for i in range(len(self._sncls)):
                if compile(self._sncls[i]).match(sncl) and overlapping_dates(self._dates[i], dates):
                    raise Exception('A pattern in %s matches an entry in %s (%s %s)' %
                                    (self._path, path, sncl, format_dates(dates)))


class RequestComparison(object):
    """
    Compare requests in two paths, checking for overlaps in requests.
    """

    def __init__(self, path1, path2):
        self._path1 = path1
        self._path2 = path2

    def _build_regexp(self, path):
        sncls_and_dates = list(parse_request(path, True))
        if sncls_and_dates:
            sncls, dates = list(zip(*sncls_and_dates))
            return Regexps(sncls, dates, path)
        else:
            return None

    def assert_no_overlap(self):
        """
        Compare the files and raise an exception on overlap.
        """
        # we need to be careful here, because the files can contain wildcards.
        # so we build regexps of the wildcards and test those, in addition to explicit equality.
        # note that we're careful to use regexp that match the regexp themselves.
        regexp1 = self._build_regexp(self._path1)
        regexp2 = self._build_regexp(self._path2)
        # we sort manually, in-memory, rather than on-disc because we don't trust the formatting to be regular
        values1 = iter(unique(parse_request(self._path1, False)))
        values2 = iter(unique(parse_request(self._path2, False)))
        # because the files are sorted we can do a fast O(n) comparison (think merge sort)
        try:
            sncl1, dates1 = next(values1)
            sncl2, dates2 = next(values2)
            while True:
                if regexp1:
                    regexp1.assert_no_overlap(sncl2, dates2, self._path2)
                if regexp2:
                    regexp2.assert_no_overlap(sncl1, dates1, self._path1)
                if sncl1 == sncl2 and overlapping_dates(dates1, dates2):
                    raise Exception('Overlap in %s and %s (%s %s and %s %s)' %
                                    (self._path1, self._path2, sncl1, format_dates(dates1), sncl2, format_dates(dates2)))
                if sncl1 < sncl2:
                    sncl1, dates1 = next(values1)
                else:
                    sncl2, dates2 = next(values2)
        except StopIteration:
            return False



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from builtins import super
from future import standard_library
standard_library.install_aliases()
from builtins import object
from sqlite3 import OperationalError
from time import sleep, time

from rover import __version__
from .args import START, DAEMON, ROVERCMD, RECHECKPERIOD, PREINDEX, POSTSUMMARY, fail_early, STOP, UserFeedback, \
    FORCECMD
from .config import write_config
from .manager import DownloadManager
from .report import Reporter
from .index import Indexer
from .process import ProcessManager
from .sqlite import SqliteSupport
from .summary import Summarizer
from .utils import check_cmd, run, windows

"""
Commands related to the daemon:

The 'rover daemon' command - equivalent to related retrieval, but in the background.
The 'rover start' command - start the daemon,
The 'rover status' command - show whether the dameon is running.
The 'rover stop' command - stop teh daemon.
"""


DAEMONCONFIG = 'rover_daemon_config'
DOWNLOADCONFIG = 'rover_download_config'


class Starter(UserFeedback):
    """
### Start

Starts the background, daemon, process to support `rover subscribe`. The
option, --recheck-period, sets the time interval in hours for the daemon to
reprocess. ROVER start is the preferred method to initiate the retrieval of
data via subscription(s).

See also `rover stop`, `rover status` and `rover daemon`.

##### Significant Options

@rover-cmd
@mseedindex-cmd
@download-workers
@mseedindex-workers
@temp-dir
@subscriptions-dir
@recheck-period
@download-retries
@http-timeout
@http-retries
@web
@http-bind-address
@http-port
@email
@email-from
@smtp-address
@smtp-port
@verbosity
@log-dir
@log-verbosity
@dev

In addition, options relevant to the processing pipeline (see `rover retrieve`
, or the individual commands for download, ingest and index) apply.

##### Examples

    rover start -f roverrc

will start the daemon using the given configuration file.

    rover start --recheck-period 24

will start the daemon, processing subscriptions every 24 hours.

    """

    def __init__(self, config):
        UserFeedback.__init__(self, config)
        fail_early(config)
        self._rover_cmd = check_cmd(config, ROVERCMD, 'rover')
        # don't clean this because it may be long-lived (it will be over-written on re-use)
        self._config_path = write_config(config, DAEMONCONFIG, verbosity=0)

    def run(self, args):
        if args:
            raise Exception('Usage: rover %s' % START)
        if windows():
            run('%s %s -f %s' % ('pythonw -m rover', DAEMON, self._config_path), self._log, uncouple=True)
        else:
            run('%s %s -f %s' % (self._rover_cmd, DAEMON, self._config_path), self._log, uncouple=True)
        self._log.default('ROVER version %s - starting %s' % (__version__, DAEMON))
        self.display_feedback()


class Stopper(object):
    """
### Stop

Stop the background, daemon, process to support `rover subscribe`.

See also `rover start`, `rover status`.

##### Significant Options

@verbosity
@log-dir
@log-verbosity

##### Examples

    rover stop -f roverrc

will stop the daemon that was started using the given configuration file.

    """

    def __init__(self, config):
        self._log = config.log
        self._processes = ProcessManager(config)

    def run(self, args):
        if args:
            raise Exception('Usage: rover %s' % STOP)
        self._processes.kill_daemon()  # logs suitable message


class StatusShower(object):
    """
### Status

Displays if the daemon is operating.

See also `rover start`, `rover stop`.

##### Significant Options

@verbosity
@log-dir
@log-verbosity

##### Examples

    rover status -f roverrc

will show whether a daemon, using a given configuration file, is running.

    """

    def __init__(self, config):
        self._log = config.log
        self._processes = ProcessManager(config)

    def run(self, args):
        if args:
            raise Exception('Usage: rover %s' % STOP)
        print()
        print(self._processes.daemon_status())
        print()




class NoSubscription(Exception):
    """
    No suitable subscription found while scanning the database.
    """


class Daemon(SqliteSupport):
    """
### Daemon

The background, daemon, process that supports `rover subscribe`. ROVER's start
command is the preferred method to launch the subscription service.

See also `rover stop`, `rover status`.

##### Significant Options

@rover-cmd
@mseedindex-cmd
@download-workers
@mseedindex-workers
@temp-dir
@subscriptions-dir
@recheck-period
@download-retries
@http-timeout
@http-retries
@web
@http-bind-address
@http-port
@email
@email-from
@smtp-address
@smtp-port
@verbosity
@log-dir
@log-verbosity
@dev

##### Examples

    rover daemon -f roverrc

will start the daemon in the foreground using the given configuration file.

    rover start --recheck-period 24

will start the daemon in the foreground, processing subscriptions every 24 hours.

REMINDER: ROVER start is the preferred method to launch the subscription service.

    """

    def __init__(self, config):
        super().__init__(config)
        self._log = config.log
        self._pre_index = config.arg(PREINDEX)
        self._post_summary = config.arg(POSTSUMMARY)
        self._download_manager = DownloadManager(config, DOWNLOADCONFIG)
        self._recheck_period = config.arg(RECHECKPERIOD) * 60 * 60
        self._reporter = Reporter(config)
        self._config = config

    def run(self, args):
        if args:
            raise Exception('Usage: rover %s' % DAEMON)
        if self._pre_index:
            Indexer(self._config).run([])
        while True:
            try:
                try:
                    id = self._find_next_subscription()
                    self._add_subscription(id)
                except NoSubscription:
                    if self._download_manager.is_idle():
                        sleep(60)
                self._download_manager.step()
                sleep(1)
            except Exception as e:
                self._reporter.send_email('ROVER Failure', self._reporter.describe_error(DAEMON, e))
                raise

    def _source_callback(self, source):
        self.execute('''UPDATE rover_subscriptions SET last_error_count = ?, consistent = ? WHERE id = ?''',
                     (source.errors.final_errors, source.consistent, source.name))
        if self._post_summary:
            Summarizer(self._config).run([])
        subject, msg = self._reporter.describe_daemon(source)
        self._reporter.send_email(subject, msg)

    def _find_next_subscription(self):
        now = time()
        found = [0]

        def callback(row):
            id = row[0]
            self._log.debug('Candidate: subscription %d' % id)
            if not found[0]:
                # todo - check this logic in test (try repeat while still downloading)
                if not self._download_manager.has_source(id):
                    found[0] = id

        try:
            self.foreachrow('''SELECT id FROM rover_subscriptions
                                  WHERE (last_check_epoch IS NULL OR last_check_epoch < ?)
                                  ORDER BY creation_epoch''', (now - self._recheck_period,),
                            callback, quiet=True)
        except OperationalError:
            pass  # no table exists, so no data

        if found[0]:
            return found[0]
        else:
            raise NoSubscription()

    def _add_subscription(self, id):
        try:
            path, availability_url, dataselect_url = self.fetchone(
                '''SELECT file, availability_url, dataselect_url FROM rover_subscriptions WHERE id = ?''', (id,))
            self._log.default('Adding subscription %d (%s, %s)' % (id, availability_url, dataselect_url))
            self._download_manager.add(id, path, True, availability_url, dataselect_url, self._source_callback)
        finally:
            self.execute('''UPDATE rover_subscriptions SET last_check_epoch = ? WHERE id = ?''', (time(), id))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library
standard_library.install_aliases()
from os.path import join, exists
from sys import version_info

if version_info[0] >= 3:
    from tempfile import TemporaryDirectory
else:
    from backports.tempfile import TemporaryDirectory

from rover.args import DEFAULT_LEAPURL, DEFAULT_LEAPEXPIRE, DEFAULT_HTTPTIMEOUT, DEFAULT_HTTPRETRIES
from rover.logs import init_log
from rover.utils import check_leap, tidy_timestamp

from .test_utils import WindowsTemp


def test_download_leap():
    with WindowsTemp(TemporaryDirectory) as dir:
        log = init_log(dir, '7M', 1, 5, 0, 'test', False, 0)[0]
        path = join(dir, 'leap.txt')
        check_leap(True, DEFAULT_LEAPEXPIRE, path, DEFAULT_LEAPURL, DEFAULT_HTTPTIMEOUT, DEFAULT_HTTPRETRIES, log)
        assert exists(path)


def assert_timestamp(log, value, target):
    tidy = tidy_timestamp(log, value)
    assert tidy == target, tidy


def test_tidy_timestamp():
    with WindowsTemp(TemporaryDirectory) as dir:
        log = init_log(dir, '7M', 1, 5, 0, 'test', False, 0)[0]
        assert_timestamp(log, '2018-7-4', '2018-07-04T00:00:00.000000')
        assert_timestamp(log, '2018-7-4T1:2:3.456', '2018-07-04T01:02:03.456000')
        assert_timestamp(log, '2018-7-4T1:3', '2018-07-04T01:03:00.000000')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import int
from builtins import str
from future import standard_library
standard_library.install_aliases()
from sys import version_info
if version_info[0] >= 3:
    from io import StringIO as buffer
else:
    from io import BytesIO as buffer
from os import unlink
from os.path import join

from rover import IndexLister
from rover.args import DATADIR

if version_info[0] >= 3:
    from tempfile import TemporaryDirectory
else:
    from backports.tempfile import TemporaryDirectory

from rover.index import Indexer
from .test_utils import find_root, ingest_and_index, WindowsTemp


def test_ingest_and_index():
    root = find_root()
    with WindowsTemp(TemporaryDirectory) as dir:
        config = ingest_and_index(dir, (join(root, 'tests', 'data'),))
        n = config.db.cursor().execute('select count(*) from tsindex').fetchone()[0]
        assert n == 36, n


def test_deleted_file():
    root = find_root()
    with WindowsTemp(TemporaryDirectory) as dir:
        config = ingest_and_index(dir, (join(root, 'tests', 'data'),))
        unlink(join(config.arg(DATADIR), 'IU', '2010', '058', 'ANMO.IU.2010.058'))
        indexer = Indexer(config)
        indexer.run([])
        n = config.db.cursor().execute('select count(*) from tsindex').fetchone()[0]
        assert n == 0, n


def run_list_index(dir, args):
    root = find_root()
    config = ingest_and_index(dir, (join(root, 'tests', 'data'),))
    stdout = buffer()
    IndexLister(config).run(args, stdout=stdout)
    stdout.seek(0)
    return stdout.read()


def assert_bad_args(dir, args, msg):
    try:
        run_list_index(dir, args)
    except Exception as e:
        assert msg in str(e), str(e)


def test_bad_begin():
    with WindowsTemp(TemporaryDirectory) as dir:
        assert_bad_args(dir, ['start=2010-1'], 'Cannot parse timestamp "2010-1"')


def test_ambiguous():
    with WindowsTemp(TemporaryDirectory) as dir:
        assert_bad_args(dir, ['s=foo'], 'Ambiguous parameter: s')


def test_two_flags():
    with WindowsTemp(TemporaryDirectory) as dir:
        assert_bad_args(dir, ['count', 'join'], 'Cannot specify multiple keys')


def test_count():
    with WindowsTemp(TemporaryDirectory) as dir:
        n = run_list_index(dir, ['count'])
        assert int(n) == 36, n



from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library
standard_library.install_aliases()
from builtins import object
from os import getpid
from socket import gethostname
from sys import version_info
from time import time

if version_info[0] >= 3:
    from email.message import EmailMessage
else:
    from email.mime.text import MIMEText
from smtplib import SMTP

from .manager import INCONSISTENT, UNCERTAIN
from .utils import format_time_epoch, format_time_epoch_local
from .args import EMAIL, EMAILFROM, SMTPPORT, SMTPADDRESS, RETRIEVE, RECHECKPERIOD, LIST_RETRIEVE, LIST_SUBSCRIBE, \
    DAEMON, TRIGGER, mm, DOWNLOADRETRIES


"""
Support for emailing the user after a download finishes.

(This is not exposed as a direct ROVER command).
"""


class Reporter(object):
    """
    Encapsulate the email functionality needed by the system.
    """

    def __init__(self, config):
        self._email_to = config.arg(EMAIL)
        self._email_from = config.arg(EMAILFROM)
        self._smtp_address = config.arg(SMTPADDRESS)
        self._smtp_port = config.arg(SMTPPORT)
        self._recheck_period = config.arg(RECHECKPERIOD)
        self._log = config.log

    def send_email(self, subject, msg):
        """
        Send the email using the pre-configured parameters.
        """
        if not self._email_to:
            self._log.info('Not sending email (see %s)' % mm(EMAIL))
        else:
            try:
                if version_info[0] >= 3:
                    email = EmailMessage()
                    email.set_content(msg)
                else:
                    email = MIMEText(msg, _subtype='plain', _charset='utf-8')
                email['Subject'] = subject
                email['From'] = self._email_from
                email['To'] = self._email_to
                self._log.default('Sending completion email to %s (subject %s)' % (self._email_to, subject))
                smtp = SMTP(self._smtp_address, port=self._smtp_port)
                if version_info[0] >= 3:
                    smtp.send_message(email)
                else:
                    smtp.sendmail(self._email_from, [self._email_to], email.as_string())
                smtp.quit()
            except Exception as e:
                self._log.error('Error sending email to %s via %s:%d: %s' %
                                (self._email_to, self._smtp_address, self._smtp_port, e))
                self._log.error('Consider using %s and %s (see your local network admin)'
                                % (mm(SMTPADDRESS), mm(EMAILFROM)))

    @staticmethod
    def _log_message(msg, logger):
        for line in msg.split('\n'):
            line = line.rstrip()
            logger(line)

    @staticmethod
    def _human_duration(seconds):
        if seconds > 86400:
            return "{0:0.1f} days".format(seconds / 86400)
        elif seconds > 3600:
            return "{0:0.1f} hours".format(seconds / 3600)
        elif seconds > 60:
            return "{0:0.1f} minutes".format(seconds / 60)
        else:
            return "{0:0.2f} seconds".format(seconds)

    @staticmethod
    def _human_size(bytes):
        if bytes < 1000:
            return "{0:d} bytes".format(bytes)
        elif (bytes / 1024) < 1000:
            return "{0:0.1f} KiB".format(bytes / 1024)
        elif (bytes / 1024 / 1024) < 1000:
            return "{0:0.1f} MiB".format(bytes / 1024 / 1024)
        elif (bytes / 1024 / 1024 / 1024) < 1000:
            return "{0:0.1f} GiB".format(bytes / 1024 / 1024 / 1024)
        elif (bytes / 1024 / 1024 / 1024 / 1024) < 1000:
            return "{0:0.1f} TiB".format(bytes / 1024 / 1024 / 1024 / 1024)
        else:
            return "{0:0.1f} PiB".format(bytes / 1024 / 1024 / 1024 / 1024 / 1024)

    def describe_error(self, task, error):
        """
        Generate an email for critical errors.
        """
        return '''
The ROVER %s (PID %d) task on %s has failed with the error:

  %s
  (%s)
''' % (task, getpid(), gethostname(), error, error.__class__.__name__)

    def describe_retrieve(self, source):
        """
        Generate the message sent by `rover retrieve`.
        """
        msg = '''
----- Retrieval Finished -----


A ROVER %s task on %s
started %s (%s UTC)
has completed in %s

The download for %d stations totaled %s,
with data covering %d seconds.

A total of %d downloads were made, with %d errors (%d on final pass of %d).
''' % (RETRIEVE, gethostname(),
       format_time_epoch_local(source.start_epoch),
       format_time_epoch(source.start_epoch),
       self._human_duration(time() - source.start_epoch),
       source.initial_progress.stations[1],
       self._human_size(source.initial_progress.download_total_bytes),
       source.initial_progress.seconds[1],
       source.errors.downloads, source.errors.errors, source.errors.final_errors,
       source.n_retries)
        if source.errors.final_errors:
            msg += '''
WARNING: The final download had some errors, it may be incomplete.
         To check for completeness use `rover %s`
         Re-run the %s command to ensure completeness.
''' % (LIST_RETRIEVE, RETRIEVE)
        elif source.consistent == INCONSISTENT:
            msg += '''
WARNING: Inconsistent retrieval was detected, most likely
         data indicated by the availability service was not
         available from the dataselect service.
'''
        elif source.consistent == UNCERTAIN:
            msg += '''
The consistency of the web services could not be confirmed.
Re-run the %s command with %s > 1 to check
''' % (RETRIEVE, mm(DOWNLOADRETRIES))
        self._log_message(msg, self._log.warn if source.errors.final_errors else self._log.default)
        return 'ROVER %s complete' % RETRIEVE, msg

    def describe_daemon(self, source):
        """
        Generate the message sent by the daemon.
        """
        msg = '''
----- Subscription Processed -----

Subscription %s has been processed by the rover %s on %s.

The task comprised of %d stations with data covering %ds.

A total of %d downloads were made, with %d errors (%d on
final pass of %d).

The subscription will be checked again in %d hours.
''' % (source.name, DAEMON, gethostname(),
       source.initial_progress.stations[1], source.initial_progress.seconds[1],
       source.errors.downloads, source.errors.errors, source.errors.final_errors,
       source.n_retries, self._recheck_period)
        if source.errors.final_errors:
            msg += '''
WARNING: The final download had some errors, it may be incomplete.
         To check for completeness use `rover %s %s`
         Run `rover %s %s` to reprocess immediately.
''' % (LIST_SUBSCRIBE, source.name, TRIGGER, source.name)
        elif source.consistent == INCONSISTENT:
            msg += '''
WARNING: Inconsistent behaviour was detected in the web
         services (eg dataselect not providing data promised
         by availability)
'''
        elif source.consistent == UNCERTAIN:
            msg += '''
The consistency of the web services could not be confirmed
(daemon must have %s > 1).
''' % mm(DOWNLOADRETRIES)
        self._log_message(msg, self._log.warn if source.errors.final_errors else self._log.default)
        return 'ROVER subscription %s processed' % source.name, msg



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import open
from builtins import str
from future import standard_library
standard_library.install_aliases()
import os
import shutil
from collections import namedtuple
from .args import STATIONURL, RETRIEVE_METADATA, UserFeedback, fail_early, \
    HTTPTIMEOUT, HTTPRETRIES, TEMPDIR, OUTPUT_FORMAT
from .report import Reporter
from .sqlite import SqliteContext, NoResult
from .utils import post_to_file, diagnose_error, unique_path, safe_unlink
from .config import timeseries_db


try:
    from io import StringIO
except ImportError:
    from io import StringIO  # python 3


TMPDOWNLOAD = 'rover_metadata_download'


class MetadataSource(UserFeedback):

    def __init__(self, config):
        UserFeedback.__init__(self, config)
        self._station_url = config.arg(STATIONURL)
        self._http_timeout = config.arg(HTTPTIMEOUT)
        self._http_retries = config.arg(HTTPRETRIES)
        self._temp_dir = config.dir(TEMPDIR)
        self._config = config

    def _do_download(self, url, in_path, out_path):
        try:
            response, check_status = post_to_file(url, in_path, out_path,
                                                  self._http_timeout,
                                                  self._http_retries,
                                                  self._log)
            check_status()
        except Exception as e:
            diagnose_error(self._log, str(e), in_path,
                           out_path, copied=False)
            raise
        return response

    def download_stationxml(self, net_code, sta_code, loc_code, cha_code,
                            starttime, endtime):
        '''
        Create station web service post request for fdsnws-station service
        and return StationXML.
        '''
        request_path = unique_path(self._temp_dir,
                                  TMPDOWNLOAD,
                                  str(os.getpid()))
        stationxml_path = unique_path(self._temp_dir,
                                       TMPDOWNLOAD,
                                       str(os.getpid()))
        ws_request = StringIO()
        ws_request.write("level=response\n"
                         "includecomments=false\n")
        ws_request.write("{} {} {} {} {} {}\n".format(net_code, sta_code,
                                                      loc_code, cha_code,
                                                      starttime, endtime))
        with open(request_path, 'w') as fd:
            ws_request.seek(0)
            shutil.copyfileobj(ws_request, fd)
        stationxml = self._do_download(self._station_url,
                                       request_path,
                                       stationxml_path)
        safe_unlink(request_path)
        return stationxml

    def has_tsindex(self):
        with SqliteContext(timeseries_db(self._config), self._log) as db:
            try:
                db.fetchone(
                    "SELECT name "
                    "FROM sqlite_master "
                    "WHERE type='table' AND name='tsindex'")
                return True
            except NoResult:
                return False

    def has_tsindex_summary(self):
        with SqliteContext(timeseries_db(self._config), self._log) as db:
            try:
                db.fetchone(
                    "SELECT name "
                    "FROM sqlite_master "
                    "WHERE type='table' AND name='tsindex_summary'")
                return True
            except NoResult:
                return False

    def fetch_summary_rows(self):
        '''
        Fetch all summary rows.

        Returns rows as list of named tuples containing:
        (network,station,location,channel,earliest,latest,updated)

        :rtype: list(tuple)
        :returns: Return rows as list of named tuples containing:
            (network, station, location, channel, earliest, latest, updated).
        '''
        NamedRow = namedtuple('NamedRow',
                              ['network', 'station', 'location', 'channel',
                               'earliest', 'latest', 'updated'])
        named_summary_rows = []
        try:
            with SqliteContext(timeseries_db(self._config), self._log) as db:
                if not self.has_tsindex():
                    raise Exception("No index - first time using rover?"
                                    .format(timeseries_db(self._config)))
                elif not self.has_tsindex_summary():
                    raise Exception("No index summary exists in database "
                                    "'{}'. Run `rover summary` to create one."
                                    .format(timeseries_db(self._config)))
                summary_rows = db.fetchall(
                        "SELECT DISTINCT s.network,s.station,s.location,"
                        "s.channel,s.earliest,s.latest,s.updt "
                        "FROM tsindex_summary s "
                        "ORDER BY s.network, s.station, s.location, "
                        "s.channel, s.earliest, s.latest")
                for row in summary_rows:
                    named_summary_rows.append(NamedRow(*row))
        except Exception as err:
            raise ValueError(str(err))

        self._log.default("Fetched %d summary rows from "
                          "tsindex_summary table." % len(named_summary_rows))
        return named_summary_rows

"""
Commands related to metadata retrieval:

The `rover retrieve-metadata` command - check for remote metadata that we
don't already have and download it.
"""

class MetadataRetriever(UserFeedback):
    """
### Retrieve-Metadata

    rover retrieve-metadata

Download missing metadata from the fdsnws-station web service and save to the
data archive. This feature is only supported for the ASDF output format.

##### Significant Options

@temp-dir
@station-url
@rover-cmd
@data-dir
@download-retries
@http-timeout
@http-retries
@http-bind-address
@http-port
@email
@email-from
@smtp-address
@smtp-port
@verbosity
@log-dir
@log-verbosity
@output-format
@asdf-filename
@force-metadata-reload

##### Examples

If the "output-format" rover.config setting is set to "asdf" then

    rover retrieve-metadata

will download missing metadata from the asdf.h5 repository.
"""

    def __init__(self, config):
        UserFeedback.__init__(self, config)
        fail_early(config)
        self._reporter = Reporter(config)
        self._config = config

    def process_asdf(self):
        """
        Retrieve metadata and insert into ASDF
        """
        from .asdf import ASDFHandler
        source = MetadataSource(self._config)
        ASDFHandler(self._config).load_metadata(source)

    def run(self, arg):
        """
        Set-up environment, parse commands, and delegate to sub-methods as
        appropriate.
        """
        try:
            self._log.default("Trying new metadata retrieval.")

            if not os.path.exists(timeseries_db(self._config)):
                raise Exception("No timeseries index database exists at {}."
                                .format(timeseries_db(self._config)))

            if self._config.arg(OUTPUT_FORMAT).upper() == "ASDF":
                self.process_asdf()
            else:
                raise Exception("Incompatible output format. "
                                "Command 'rover retrieve-metadata' "
                                "can only be used with 'output-format=asdf'.")
            self._log.default("")
            self._log.default("----- Metadata Retrieval Finished -----")
            self._log.default("")
        except Exception as e:
            self._reporter.send_email('Rover Failure',
                                      self._reporter.describe_error(
                                                        RETRIEVE_METADATA, e))
            raise



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from builtins import dict
from future import standard_library
standard_library.install_aliases()
from builtins import object
from fnmatch import fnmatch
from getpass import getuser
from os import listdir, makedirs
from os.path import dirname, join, exists
from random import randint
from re import sub
from shutil import rmtree
from time import time

import rover
from rover.config import BaseConfig
from rover.args import Arguments, DATADIR, TEMPDIR, LOGDIR, LEAP, MSEEDINDEXCMD
from rover.ingest import Ingester
from rover.logs import init_log
from rover.sqlite import init_db
from rover.utils import create_parents, canonify, windows


class TestArgs(object):

    def __init__(self, **kargs):
        self._kargs = kargs
        self._argparser = Arguments()

    def __getattr__(self, item):
        if item in self._kargs:
            value = self._kargs[item]
        else:
            value = self._argparser.get_default(item)
        return value


def _(text):
    return sub(r'\-', '_', text)


class TestConfig(BaseConfig):

    def __init__(self, dir, **kargs):
        kargs = dict(kargs)
        kargs[_(DATADIR)] = join(dir, 'data')
        kargs[_(TEMPDIR)] = join(dir, 'tmp')
        kargs[_(LOGDIR)] = join(dir, 'logs')
        root = find_root()
        kargs[_(MSEEDINDEXCMD)] = join(root, '..', 'mseedindex', 'mseedindex')
        kargs[_(LEAP)] = False
        args = TestArgs(**kargs)
        self.command = args.command
        log, log_path, log_stream = init_log(args.log_dir, '7M', 1, 5, 0, 'test', args.leap, 0)
        dbpath = join(canonify(args.data_dir), 'timeseries.sqlite')
        create_parents(dbpath)
        super().__init__(log, log_path, args, init_db(dbpath, log), dir)


def find_root():
    return dirname(dirname(rover.__file__))


def assert_files(dir, *files):
    found = listdir(dir)
    if 'timeseries.sqlite' in found:
        found.remove('timeseries.sqlite')
    assert len(files) == len(found), 'Found %d files in %s (not %d)' % (len(found), dir, len(files))
    for file in found:
        ok = False
        for glob in files:
            if fnmatch(file, glob):
                ok = True
        if not ok:
            raise Exception('Unexpected file: "%s"', file)


def ingest_and_index(dir, data, **opts):
    config = TestConfig(dir, **opts)
    ingester = Ingester(config)
    ingester.run(data)
    return config


class WindowsTemp(object):
    """
    An ugly hack to work around problems with TemporaryDir on windows.
    """

    def __init__(self, context, cleanup=True):
        self._context = context
        self._cleanup = cleanup
        self._dir = "C:\\Users\\%s\\AppData\\Local\\Temp\\rover\\%d-%s" % (getuser(), time(), randint(1000, 9999))

    def __enter__(self):
        if windows():
            if not exists(self._dir):
                makedirs(self._dir)
            return self._dir
        else:
            self._context = self._context()
            return self._context.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if windows():
            if self._cleanup:
                rmtree(self._dir, ignore_errors=True)
        else:
            return self._context.__exit__(exc_type, exc_val, exc_tb)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library
standard_library.install_aliases()
from sys import version_info

from rover import ProcessManager

if version_info[0] >= 3:
    from tempfile import TemporaryDirectory
else:
    from backports.tempfile import TemporaryDirectory

from rover.retrieve import ListRetriever, Retriever
from .test_utils import TestConfig, WindowsTemp


def test_retrieve():
    with WindowsTemp(TemporaryDirectory, False) as dir:
        config = TestConfig(dir, args=['IU_ANMO_00_BH1', '2017-01-01', '2017-01-04'], command='retrieve',
                            rover_cmd='python -m rover', verbosity=0, delete_files=False, dev=True, log_verbosity=5)
        n_sncls = ListRetriever(config).run(config.args)
        assert n_sncls == 1, n_sncls
        with ProcessManager(config):
            n_downloads = Retriever(config).run(config.args)
        assert n_downloads == 4, n_downloads
        n_sncls = ListRetriever(config).run(config.args)
        assert n_sncls == 0, n_sncls
        # alternate syntax
        config = TestConfig(dir, args=['net=IU', 'sta=ANMO', 'location=00', 'c=BH1', '2017-01-01', '2017-01-04'],
                            rover_cmd='python -m rover', verbosity=0, delete_files=False, dev=True, log_verbosity=5)
        n_sncls = ListRetriever(config).run(config.args)
        assert n_sncls == 0, n_sncls



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import open
from future import standard_library
standard_library.install_aliases()
from builtins import object
import json
import os

from subprocess import Popen
from time import sleep

from .args import TEMPDIR
from .utils import uniqueish, unique_filename

"""
Support for running multiple sub-processes.
"""

class Workers(object):
    """
    A collection of processes that run asynchronously.  Note that the Python
    code here does NOT run asynchonously - it will block if there are no free
    workers.

    The idea is that (for example) we run N mseedindex processes in the background
    so that we are not waiting on them to complete.
    """

    def __init__(self, config, n_workers):
        self._temp_dir = config.dir(TEMPDIR)
        self._log = config.log
        self._n_workers = n_workers
        self._workers = []  # (command, popen, callback)

    def execute(self, command, callback=None, feedback=None):
        """
        Execute the command in a separate process.
        """
        self._wait_for_space()

        if not callback:
            callback = self._default_callback

        if feedback:
            name = uniqueish('rover_worker_feedback', command)
            filename = unique_filename(os.path.join(self._temp_dir, name))

            try:
                feedback = open (filename, 'w+')
            except Exception as ex:
                raise Exception('Cannot open feedback file: %s' % ex)

        self._log.debug('Adding worker for "%s" (callback %s)' % (command, callback))
        self._workers.append((command, self._popen(command, feedback=feedback), callback, feedback))

    def _wait_for_space(self):
        while True:
            self.check()
            if self.has_space():
                self._log.debug('Space for new worker (%d/%d)' % (len(self._workers), self._n_workers))
                return
            sleep(0.1)

    def has_space(self):
        return len(self._workers) < self._n_workers

    def _default_callback(self, cmd, returncode, **kwargs):
        if returncode:
            raise Exception('"%s" returned %d' % (cmd, returncode))
        else:
            self._log.debug('"%s" succeeded' % (cmd,))

    def check(self):
        for idx, worker in enumerate(self._workers):
            command, process, callback, feedback = worker

            process.poll()
            if process.returncode is not None:

                # Remove finished worker from list
                del self._workers[idx]

                self._log.debug('Calling callback %s (command %s)' % (callback, command))

                process_feedback = {}
                if feedback:
                    try:
                        feedback.seek(0, 0)
                        feedback_data = feedback.read()

                        if len(feedback_data) > 0:
                            process_feedback = json.loads(feedback_data)
                    except Exception as ex:
                        self._log.error('Error processing feedback file: %s, contents: %s' % (ex, feedback_data))
                    finally:
                        feedback.close()
                        os.remove(feedback.name)

                if process_feedback:
                    callback(command, process.returncode, feedback=process_feedback)
                else:
                    callback(command, process.returncode)

    def wait_for_all(self):
        """
        Wait for all remaining processes to finish.
        """
        while True:
            self.check()
            if not self._workers:
                self._log.debug('No workers remain')
                return
            sleep(0.1)

    def _popen(self, command, feedback=None):
        return Popen(command, shell=True, stdout=feedback)



from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import super
from builtins import next
from builtins import map
from builtins import open
from builtins import str
from builtins import int
from future import standard_library
standard_library.install_aliases()
from builtins import object
import datetime as dt
from collections import deque
from random import randint
from sqlite3 import OperationalError
from time import time, sleep

from .args import mm, FORCEFAILURES, DELETEFILES, TEMPDIR, HTTPTIMEOUT, HTTPRETRIES, TIMESPANTOL, DOWNLOADRETRIES, \
    DOWNLOADWORKERS, ROVERCMD, MSEEDINDEXCMD, LOGUNIQUE, LOGVERBOSITY, VERBOSITY, DOWNLOAD, DEV, WEB, SORTINPYTHON, \
    TIMESPANINC, ABORT_CODE
from .config import write_config
from .coverage import Coverage, SingleSNCLBuilder
from .download import DEFAULT_NAME, TMPREQUEST, TMPRESPONSE
from .sqlite import SqliteSupport
from .utils import utc, EPOCH_UTC, PushBackIterator, format_epoch, safe_unlink, unique_path, post_to_file, \
    sort_file_inplace, parse_epoch, check_cmd, run, windows, diagnose_error, format_year_day_epoch
from .workers import Workers

"""
The core logic for scheduling multiple downloads.  Called by both the daemon and `rover retrieve`.
"""


class ManagerException(Exception):
    """
    Separate class so we can avoid sending additional emails when manager fails.
    """


class ProgressStatistics(object):
    """
    Encapsulate the statistics that measure progress for a single retrieval.
    Statistics are pairs of values:
      index 0 - current value
      index 1 - initial value / limit
    """

    download_bytes = 0 # number of downloaded bytes per run
    download_total_bytes = 0 # total number of downloaded bytes
    download_retry_count = 1 # number of times that tried to download

    def __init__(self):
        self.__prev_net_sta = [None, None]
        self.stations = [0, 0]
        self.seconds = [0, 0]
        self.chunks = [0, 0]

    def add_coverage(self, coverage):
        net_sta = coverage.sncl.split('_')[0:2]
        if net_sta != self.__prev_net_sta:
            self.stations[1] += 1
            self.__prev_net_sta = net_sta
        # but seconds are counted either way
        for start, end in coverage.timespans:
            self.seconds[1] += (end - start)

    def pop_timespan(self, start, end):
        self.seconds[0] += (end - start)

    def add_chunks(self, n):
        self.stations[0] += 1  # a set of chunks is for a single station
        # the day count isn't global - it's per sncl (coverage) - so resets
        self.chunks[0] = 0
        self.chunks[1] = n

    def pop_chunk(self):
        self.chunks[0] += 1

    def __str__(self):
        return '(N_S %d/%d; day %d/%d)' % (self.stations[0], self.stations[1], self.chunks[0], self.chunks[1])


class ErrorStatistics(object):
    """
    Encapsulate the statistics that count errors.
    Raw values are per retrieval, they are accumulated on the source.
    """
    def __init__(self):
        self.downloads = 0
        self.errors = 0
        self.final_errors = None

    def accumulate(self, errors):
        self.downloads += errors.downloads
        self.errors += errors.errors
        self.final_errors = errors.errors


class Chunks(object):
    """
    A chunk is a collection of SNCLs and timespans that are downloaded at once.
    We used to download each individually, but that required too many parallel requests to
    be efficient, so now we collect them here.
    This is a *collection* of chunks because a single chunk is only for one calendar day -
    we may assemble multiple days when moving from the coverages to chunks.
    """

    def __init__(self, temp_dir):
        self.__temp_dir = temp_dir
        self.__chunks = {}   # list of (sncl, start, end) indexed by end of day epoch
        self.__network = None
        self.__station = None

    def __bool__(self):
        return bool(self.__chunks)

    def __len__(self):
        return len(self.__chunks)

    @staticmethod
    def _end_of_day(epoch):
        day = dt.datetime.fromtimestamp(epoch, utc)
        right = (dt.datetime(day.year, day.month, day.day, tzinfo=utc)
                 + dt.timedelta(hours=24) - EPOCH_UTC).total_seconds()
        left = right - 0.000001
        return left, right

    def _append(self, right, sncl, start, end):
        if right not in self.__chunks:
            self.__chunks[right] = []
        self.__chunks[right].append((sncl, start, end))

    def _set_ns(self, sncl):
        nslc = sncl.split('_')
        self.__network = nslc[0]
        self.__station = nslc[1]

    def sncl_ok(self, sncl):
        nslc = sncl.split('_')
        return not self.__chunks or (nslc[0] == self.__network and nslc[1] == self.__station)

    def add_coverage(self, coverage):
        sncl, timespans = coverage.sncl, PushBackIterator(iter(coverage.timespans))
        if not self.__chunks:
            self._set_ns(sncl)

        # Determine sampling period (interval)
        # On initial download we do not know the sampling rate/period, but if data exists locally we do
        # This indicates that data bounds are determined from local data
        try:
            sampleperiod = 1 / coverage.samplerate
        except:
            sampleperiod = None

        for start, end in timespans:
            left, right = self._end_of_day(start)

            # If end time is before end of day, append whole range.
            if right > end:
                # Skip if max request would be smaller than sample period, if known (implying local data bounds)
                # In this case, assuming the sampling interval is regular, no data is expected
                if sampleperiod and sampleperiod > 0 and (left - start) < sampleperiod:
                    continue
                else:
                    self._append(right, sncl, start, end)

            # Otherwise, add the range beyond the current day to the timespans and
            # append the range that fits in the first day
            else:
                timespans.push((right, max(end, right)))

                # Skip if max request would be smaller than sample period, if known (implying local data bounds)
                # In this case, assuming the sampling interval is regular, no data is expected
                if sampleperiod and sampleperiod > 0 and (left - start) < sampleperiod:
                    continue
                else:
                    self._append(right, sncl, start, left)


    @staticmethod
    def format_sncl(sncl):
        return ' '.join(code if code else '--' for code in sncl.split('_'))

    def pop(self, progress):
        right = next(iter(sorted(self.__chunks.keys())))
        description = '%s_%s %s' % (self.__network, self.__station, format_year_day_epoch(right-24*3600))
        data = self.__chunks[right]
        path = unique_path(self.__temp_dir, 'rover_chunk', description)
        with open(path, 'w') as out:
            for (sncl, start, end) in data:
                progress.pop_timespan(start, end)
                print('%s %s %s' % (self.format_sncl(sncl), format_epoch(start), format_epoch(end)), file=out)
        del self.__chunks[right]
        progress.pop_chunk()
        return description, path


class Retrieval(object):
    """
    A single attempt at downloading data for a subscription or retrieval
    (some care is taken here to avoid large files in memory - the list of downloads is generated
    on the fly).
    """

    def __init__(self, log, name, temp_dir, delete_files, dataselect_url, force_failures):
        self._log = log
        self._name = name
        self._temp_dir = temp_dir
        self._delete_files = delete_files
        self._dataselect_url = dataselect_url
        self._force_failures = force_failures
        self._coverages = deque()  # fifo: appendright / popleft; exposed for display
        self._chunks = None
        self.worker_count = 0
        self.errors = ErrorStatistics()
        self.progress = ProgressStatistics()

    def add_coverage(self, coverage):
        """
        Add a coverage to retrieve.
        """
        if coverage:  # this avoids us showing more stations to download than are downloaded
            self._coverages.append(coverage)
            self.progress.add_coverage(coverage)

    def get_coverages(self):
        """
        Provide access to coverages (for listing)
        """
        return list(self._coverages)

    def has_chunks(self):
        """
        Ensure chunks has some data, if possible, and return whether it has any.
        """
        if self._chunks:
            return True
        self._chunks = Chunks(self._temp_dir)
        while self._coverages and self._chunks.sncl_ok(self._coverages[0].sncl):
            self._chunks.add_coverage(self._coverages.popleft())
        if self._chunks:
            self.progress.add_chunks(len(self._chunks))
            return True
        return False

    def _build_url(self, sncl, start, end):
        url_params = 'net=%s&sta=%s&loc=%s&cha=%s' % \
                     tuple(code if code else '--' for code in tuple(sncl.split('_')))
        return '%s?%s&start=%s&end=%s' % (self._dataselect_url, url_params, format_epoch(start), format_epoch(end))

    def _worker_callback(self, command, return_code, path, **kwargs):
        feedback = kwargs.get("feedback")
        if feedback:
            bytecount = feedback.get("download_byte_count", 0)
            ProgressStatistics.download_bytes += bytecount
            ProgressStatistics.download_total_bytes += bytecount
        if self._delete_files:
            safe_unlink(path)
        self.worker_count -= 1
        self.errors.downloads += 1
        if return_code:
            self.errors.errors += 1
            if return_code != ABORT_CODE:   # hide message on ctrl-C as we will exit as well
                self._log.error('Download %s failed (return code %d)' % (self._name, return_code))

    def new_worker(self, workers, config_path, rover_cmd):
        """
        Launch a new worker (called by manager main loop).
        """
        description, path = self._chunks.pop(self.progress)
        self._log.default('Downloading %s %s' % (description, self.progress))
        # for testing error handling we can inject random errors here
        if randint(1, 100) <= self._force_failures:
            self._log.warn('Random failure expected (%s %d)' % (mm(FORCEFAILURES), self._force_failures))
            command = 'exit 1  # failure for tests'
        else:
            # we only pass arguments on the command line that are different from the
            # default (which is in the file)
            if windows():
                command = 'pythonw -m rover -f %s %s "%s"' % (config_path, DOWNLOAD, path)
            else:
                command = '%s -f %s %s "%s"' % (rover_cmd, config_path, DOWNLOAD, path)
        self._log.debug(command)

        callback_function = lambda cmd, rtn, **kwargs: self._worker_callback(cmd, rtn, path, **kwargs)

        try:
            workers.execute(command, callback=callback_function, feedback=True)
        except Exception as ex:
            self._log.error('Worker failed (%s): %s' % (command, ex))
        else:
            self.worker_count += 1

    def is_complete(self):
        """
        Is this retrieval complete?
        """
        return self.worker_count == 0 and not self.has_chunks()


# avoid enum because python2 doesn't have it and we want code that runs on both
# (if we use backports then it's a conditional install)
UNCERTAIN, CONFIRMED, INCONSISTENT = 0, 1, 2


class Source(SqliteSupport):
    """
    Data for a single source in the download manager.
    """

    # this has two jobs, really.
    # first, it runs retrievers as often as required.  this is the "main loop" logic from the
    # rfq and was added later in the project (before we failed and reported).
    # second. it collects and reports statistics that are used by the download manager and displayed to the user.
    # these are the public attributes and properties (delegated to the current retriever).

    def __init__(self, config, name, fetch, request_path, availability_url, dataselect_url, completion_callback):
        super().__init__(config)
        self._log = config.log
        self._force_failures = config.arg(FORCEFAILURES)
        self._delete_files = config.arg(DELETEFILES)
        self._temp_dir = config.dir(TEMPDIR)
        self._http_timeout = config.arg(HTTPTIMEOUT)
        self._http_retries = config.arg(HTTPRETRIES)
        self._timespan_inc = config.arg(TIMESPANINC)
        self._timespan_tol = config.arg(TIMESPANTOL)
        self._config = config
        self.download_retries = config.arg(DOWNLOADRETRIES)
        self._sort_in_python = config.arg(SORTINPYTHON)
        self.name = name
        self._request_path = request_path
        self._availability_url = availability_url
        self._dataselect_url = dataselect_url
        self._completion_callback = completion_callback
        self.n_retries = 0
        self._retrieval = None
        self.start_epoch = time()
        self.errors = ErrorStatistics()
        self._expect_empty = False
        self.consistent = UNCERTAIN
        # load first retrieval immediately so we don't print messages in the middle of list-retrieve
        self._new_retrieval(fetch)
        self.initial_progress = self._retrieval.progress

    def __str__(self):
        return '%s (%s)' % (self.name, self._dataselect_url)

    def get_coverages(self):
        """
        Provide access to coverages (for listing)
        """
        return self._retrieval.get_coverages()

    def stats(self):
        """
        Provide access to latest stats (for reporting).
        """
        return self._retrieval.progress

    def has_chunks(self):
        """
        Does retrieval have data?
        """
        return self._retrieval.has_chunks()

    @property
    def worker_count(self):
        """
        Number of workers being used.
        """
        return self._retrieval.worker_count

    def new_worker(self, workers, config_path, rover_cmd):
        """
        Launch a new worker (called by manager main loop).
        """
        self._retrieval.new_worker(workers, config_path, rover_cmd)

    @property
    def _name(self):
        """
        Name for log messages - empty string for retrieve, otherwise identifies subscription.
        """
        # don't mention subscriptions when called from `rover retrieve`
        return '' if self.name == DEFAULT_NAME else 'subscription %s ' % self.name

    def is_complete(self):
        """
        Is this source complete (all retrievals)?
        Dependent Functions:
        
        is_complete_final_read: Verifies that all data is downloaded with no errors makes ROVER robust. 
        If data is missing _new_retrieval and _is_complete_initial_reads are called.
        _is_complete_initial_reads: Call the new_retrieval function until all data is downloaded with no errors. 
        _new_retrieval: 
        
        important variables: 

        self._expect_empty: Determines if the _is_complete_final_read or _is_complete_initial_reads function is called/
        complete: Terminates the loop.
        """
        retry_possible = self.n_retries < self.download_retries

        # Only one possible rety needs to be written in the intial loop.
        # The second loop is just there to verfy the first loop.
        # The total possible retries need to counted upward in the first and second loop

        # this is complicated by the fact that we also check for consistency.
        # inconsistency is when we expect to download no data, but still get some, or expect to get
        # some, but get nothing.  we can't detect all cases, but we do our best by waiting until
        # we get a clean download (no errors and so presumably complete) and then running an
        # extra download.  that should generate no downloads.  if it doesn't, we have problems.

        # the default value for consistent is UNCERTAIN so it is left unchanged in many places below
        # we throw an exception if we finish with incomplete data (errors) or proof of inconsistency.
        if self._retrieval.is_complete():
            self.errors.accumulate(self._retrieval.errors)
            complete = True  # default if exception thrown
            try:

                if self._expect_empty:
                    complete = self._is_complete_final_read(retry_possible)
                else:
                    complete = self._is_complete_initial_reads(retry_possible)
                return complete
            finally:
                if complete:
                    self._completion_callback(self)
        else:
            # the current retrieval isn't complete, so we're certainly not done
            return False

    def _is_complete_initial_reads(self, retry_possible):

        # the last retrieval had errors.
        if self._retrieval.errors.errors:

            # if we can retry, then do so
            if retry_possible:
                self._log.default(('Retrieval attempt %d of %d completed with %d errors. '+
                                    'We will retry to check that all data were retrieved') %
                                    (self.n_retries, self.download_retries, self._retrieval.errors.errors))
                self._new_retrieval(True)
                return False
            # otherwise, we can't retry so we're done, but failed.
            else:
                raise ManagerException('Retrieval attempt %d of %d had %d errors on the final attempt.' %
                                       (self.n_retries, self.download_retries, self._retrieval.errors.errors))

        # No errors last retrieval, but we did attempt some downloads
        # If these boolean conditions are maintained then rover will continue to make download attmepts.
        elif self._retrieval.errors.downloads:
            # Can we try again, to make sure there are no more data retrievable?
            # Deal with the case when config has number of download attempts set to 1.
            if self.download_retries == 1:
                if retry_possible:

                    self._log.default(('Retrieval attempt %d of %d had no errors and we downloaded data.'
                                       'The config parameter download-retries is set to %d attempt so rover is exiting.'
                                       'We recommend increasing the number of download attempts. ')
                                      % (self.n_retries, self.download_retries, self.download_retries))
                    self._expect_empty = True
                    return True
            else:
                if ProgressStatistics.download_bytes > 0: # Determines if data was collected.
                    if ProgressStatistics.download_retry_count < 100: # A hard limit on the amount of times we will attempt to download a request
                        self._log.default('Successful retrieval, downloaded data so resetting retry count and verify.')
                        self.n_retries = self.n_retries - 1
                        self._expect_empty = False
                        self._new_retrieval(True)
                        ProgressStatistics.download_bytes = 0
                        ProgressStatistics.download_retry_count += 1
                        return False
                    else :
                        self._log.default(('The retrieval has been attempted %d times. ROVER is exiting. Please check the log files and try again later.')
                                          % (ProgressStatistics.download_retry_count))
                        self._expect_empty =True
                        return True
                elif retry_possible:
                    self._log.default(('Successful retrieval attempt %d of %d, but no new data downloaded so try again.')
                                      % (self.n_retries, self.download_retries))

                    self._expect_empty =False
                    self._new_retrieval(True)
                    return False
                # if not, we're going to say we're complete anyway, since we didn't have any errors.
                else:
                    self._log.default(
                        ('The latest retrieval attempt had no errors.  Stopping at %d retry attempts.') % (self.n_retries))
                    self._expect_empty =True
                    return False

        # no errors and no data. used for rovers first download attempt and to exit out the
        # _is_complete_initial_reads program so the _is_complete_final_reads program can be run.
        # When escaping is_complete_initial_reads an additional _new_retrieval should not be run.
        else:
            # presumably this was an empty initial download
            if self.n_retries == 1:
                # can we try again to make sure things are consistent?
                if retry_possible:
                    self._log.default('The initial retrieval attempt resulted in no errors or data downloaded, will verify.')
                    self._expect_empty = True
                    self._new_retrieval(True)

                    return False
                # if not, we're going to say we're complete anyway.
                else:
                    self._log.default('The initial retrieval attempt had no errors and no retries are configured.')
                    return True
            # something odd has happened - no data when it wasn't expected
            else:
                # can we try again to make sure things are consistent?
                if retry_possible:
                    self._log.default(('Retrieval attempt %d of %d had no errors and downloaded no data. ' +
                                       'We will check that all data were retrieved.') % (self.n_retries, self.download_retries))
                    self._expect_empty = True
                    return False
                else:
                    self.consistent = INCONSISTENT
                    raise ManagerException(('The final retrieval, attempt %d of %d, downloaded no data' +
                                            'following an earlier error (inconsistent web services?)') %
                                           (self.n_retries, self.download_retries))

    def _is_complete_final_read(self, retry_possible):

        # the last retrieval had errors (it shouldn't have - we should be on final empty download)
        if self._retrieval.errors.errors:
            self.consistent = INCONSISTENT
            # if we can retry, then do so
            if retry_possible:
                self._log.default(('The latest retrieval attempt had %d errors after %d attempts.  ' +
                                   'We will retry to complete the retrieval.') %
                                  (self._retrieval.errors.errors, self.n_retries))
                self._new_retrieval(True)
                self._expect_empty = False
                return False
            # otherwise, we can't retry so we're done, but failed.
            else:
                raise ManagerException('The latest retrieval attempt had %d errors on final attempt (%d of %d).' %
                                       (self._retrieval.errors.errors, self.n_retries, self.download_retries))

        # no errors last retrieval, but we did download some more data (again, unexpected)
        elif self._retrieval.errors.downloads:
            # special case where we had no sample interval until first request, so second request
            # got a single point that was missed because the interval was too small for the server
            if self.n_retries == 2:
                if retry_possible:
                    self._log.default(('The latest retrieval attempt downloaded additional data - ' +
                                       'probably an isolated point that was not retrieved on the first pass. ' +
                                       'We will retry to make sure data are complete.'))
                    self._new_retrieval(True)
                    self._expect_empty = False
                    return False
                else:
                    self._log.default(('The latest retrieval attempt downloaded additional data - ' +
                                       'probably an isolated point that was not retrieved on the first pass. ' +
                                       'We cannot be certain download is complete without more retries.'))
                    return True
            else:
                self.consistent = INCONSISTENT
                # can we try again, in case this was some weird hiccup?
                # there's a case where this is expected - when we have data on a day boundary and need tp
                # download once to get the samplerate so we can judge exactly how small a chunk to take
                # from the next day.
                if retry_possible:
                    self._log.default(('The latest retrieval attempt downloaded unexpected data so trying again ' +
                                       'to check behavior.'))
                    self._new_retrieval(True)
                    self._expect_empty = False
                    return False
                # something odd is happening
                else:
                    raise ManagerException(('The previous %d retrieval attempts could not download some '+
                                            'requested data available at the data center. Consider running'+
                                            ' `rover list-retrieve` to determine missing data.') %
                                           (self.n_retries))

        # no errors and no data
        else:
            self.consistent = CONFIRMED
            self._log.default('The final %sretrieval, attempt %d of %d, made no downloads and had no errors, we are complete.' %
                              (self._name, self.n_retries, self.download_retries))
            return True

    def _new_retrieval(self, fetch):
        # fetch indicates we're not simply querying and so should check for no data and prime days
        self.n_retries += 1
        if fetch:
            self._log.default('Trying new %sretrieval attempt %d of %d.' %
                              (self._name, self.n_retries, self.download_retries))
        self._retrieval = Retrieval(self._log, self._name, self._temp_dir, self._delete_files,
                                    self._dataselect_url, self._force_failures)
        request = self._build_request(self._request_path)
        response = self._get_availability(request, self._availability_url)
        try:
            # compare database and availability to construct list of missing data
            # we could make this lazy, but then we lose progression statistics.  so
            # just try to be as meagre with memory use as possible.
            for remote in self._parse_availability(response):
                self._log.debug('Available data: %s' % remote)
                local = self._scan_index(remote.sncl)
                self._log.debug('Local data: %s' % local)
                required = remote.subtract(local)
                self._retrieval.add_coverage(required)
        finally:
            if self._delete_files:
                safe_unlink(request)
                safe_unlink(response)
        if fetch and not self._retrieval.has_chunks():
            self._log.default('%sRetrieval attempt %d of %d is complete.' %
                              (self._name, self.n_retries, self.download_retries))

    def _build_request(self, path):
        tmp = unique_path(self._temp_dir, TMPREQUEST, path)
        self._log.debug('Prepending options to %s via %s' % (path, tmp))
        with open(tmp, 'w') as output:
            print('merge=samplerate,quality', file=output)
            with open(path, 'r') as inpath:
                for line in inpath:
                    print(line, file=output, end='')
        return tmp

    def _get_availability(self, request, availability_url):
        self._log.info('Checking availability service')
        response = unique_path(self._temp_dir, TMPRESPONSE, request)
        response, check_status = post_to_file(availability_url, request, response, self._http_timeout, self._http_retries, self._log)
        try:
            check_status()
            return response
        except Exception as e:
            diagnose_error(self._log, str(e), request, response)
            raise

    def _parse_line(self, line):
        try:
            n, s, l, c, b, e = ('' if token == '--' else token for token in line.split())
            return "%s_%s_%s_%s" % (n, s, l, c), parse_epoch(b), parse_epoch(e)
        except:
            raise Exception('Could not parse "%s" in the response from the availability service' % line)

    def _parse_availability(self, response):
        try:
            if response is not None:  # None when no data returned
                sort_file_inplace(self._log, response, self._temp_dir, self._sort_in_python)
                with open(response, 'r') as input:
                    availability = None
                    for line in input:
                        line = line.strip()
                        if line and not line.startswith('#'):
                            sncl, b, e = self._parse_line(line)
                            if availability and not availability.sncl == sncl:
                                yield availability
                                availability = None
                            if not availability:
                                availability = Coverage(self._log, self._timespan_tol, self._timespan_inc, sncl)
                            availability.add_epochs(b, e)
                    if availability:
                        yield availability
        except Exception as e:
            diagnose_error(self._log, 'Problems parsing the availability service response.',
                           self._request_path, response)
            raise

    def _scan_index(self, sncl):
        availability = SingleSNCLBuilder(self._log, self._timespan_tol, self._timespan_inc, sncl)

        def callback(row):
            availability.add_timespans(row[0], row[1])

        try:
            # coalesce below replaces [...] with <...> based on start/endtime if timespans is missing
            # this is handled by rover.coverage.BaseBuilder
            # see issue 47
            # note we separate times with space as time contains colons
            self.foreachrow('''SELECT coalesce(timespans, '<' || starttime || ' ' || endtime || '>'), samplerate
                                    FROM tsindex
                                    WHERE network=? AND station=? AND location=? AND channel=?
                                    ORDER BY starttime, endtime''',
                            sncl.split('_'),
                            callback, quiet=True)
        except OperationalError:
            self._log.debug('No index - check rover.config')
        return availability.coverage()


class DownloadManager(SqliteSupport):
    """
    An interface to downloader instances that restricts downloads to a fixed number of workers,
    each downloading data that is for a maximum duration of a day.

    It supports multiple *sources* and will try to divide load fairly between sources.  A
    source is typically a source / subscription, so we spread downloads across multiple
    servers when possible.

    The config_file is overwritten (in temp_dir) because only a singleton (for either
    standalone or daemon) should ever exist.  Because of this, and the daemon exiting via
    kill(), no attempt is made to delete the file on exit.

    IMPORTANT: This is used from a SINGLE thread.  So for it to work reliably the step()
    method must be called regularly (perhaps via download()).
    """

    def __init__(self, config, config_file=None):
        super().__init__(config)
        self._log = config.log
        self._config = config
        self._sources = {}  # map of source names to sources
        self._index = 0  # used to round-robin sources
        self._workers = Workers(config, config.arg(DOWNLOADWORKERS))
        self._n_downloads = 0
        self._create_stats_table()
        if config_file:
            # these aren't used to list subscriptions (when config_file is None)
            self._rover_cmd = check_cmd(config, ROVERCMD, 'rover')
            self._mseed_cmd = check_cmd(config, MSEEDINDEXCMD, 'mseedindex')
            log_unique = config.arg(LOGUNIQUE) or not config.arg(DEV)
            log_verbosity = config.arg(LOGVERBOSITY) if config.arg(DEV) else min(config.arg(LOGVERBOSITY), 3)
            self._config_path = write_config(config, config_file, log_unique=log_unique, log_verbosity=log_verbosity)
            self._start_web()
        else:
            self._config_path = None

    # source management

    def has_source(self, name):
        """
        Is the given source known (they are deleted once all data are downloaded).
        """
        return name in self._sources

    def _source(self, name):
        if name not in self._sources:
            raise Exception('Unexpected source: %s' % name)
        return self._sources[name]

    def add(self, name, request_path, fetch, availability_url, dataselect_url, completion_callback):
        # fetch is necessary here because source wants to prime days for retrieval
        if name in self._sources and self._sources[name].worker_count:
            raise Exception('Cannot overwrite active source %s' % self._sources[name])
        self._sources[name] = Source(self._config, name, fetch, request_path, availability_url, dataselect_url,
                                     completion_callback)

    # display expected downloads

    def display(self):
        """
        Display a summary of the data that have not been expanded into downloads.
        """
        total_seconds, total_sncls = 0, 0
        print()
        for name in self._sources.keys():
            source = self._sources[name]
            coverages = source.get_coverages()
            if name != DEFAULT_NAME:
                print('  Subscription %s ' % source)
                print()
            source_seconds, source_sncls = 0, 0
            for coverage in coverages:
                sncl_seconds = 0
                for (start, end) in coverage.timespans:
                    seconds = end - start
                    sncl_seconds += seconds
                    source_seconds += seconds
                    total_seconds += seconds
                source_sncls += 1
                total_sncls += 1
                print('  %s  (%4.2f sec)' % (coverage.sncl, sncl_seconds))
                for (start, end) in coverage.timespans:
                    print('    %s - %s  (%4.2f sec)' % (format_epoch(start), format_epoch(end), end - start))
            if name != DEFAULT_NAME:
                if source_sncls:
                    print()
                print('  %s: %d N_S_L_C; %4.2f sec' % (name, source_sncls, source_seconds))
                print()
        print('  Total: %d N_S_L_C; %4.2f sec' % (total_sncls, total_seconds))
        print()
        return total_sncls

    # downloading data and processing in the pipeline

    def _has_data(self):
        for source in self._sources.values():
            if source.has_chunks():
                return True
        return False

    def _next_source(self, sources):
        self._index = (self._index + 1) % len(sources)
        return sources[self._index]

    def _has_least_workers(self, c):
        for source in self._sources.values():
            if source.worker_count < c.worker_count:
                return False
        return True

    def _clean_sources(self, quiet=True):
        names = list(self._sources.keys())
        for name in names:
            try:
                complete = self._source(name).is_complete()
            except Exception as e:
                if quiet:
                    complete = True
                else:
                    raise
            if complete:
                self._log.debug('Source %s complete' % self._source(name))
                del self._sources[name]

    def is_idle(self):
        """
        Are we no longer downloading data?
        """
        self._clean_sources()
        if not self._sources:
            self._update_stats()  # wipe
            return True
        else:
            return False

    # the step method below is probably the best place to insert a call to execute
    # pragma wal_checkpoint if we have further problems with locked databases.
    # the idea would be to add a counter to the manager and every 1000 (say) calls,
    # wait for all workers to close and then execute the pragma.  this would
    # allow the database to update without contention.

    # we would also need to disable automatic checkpointing with pragma
    # wal_autocheckpoint and make sure to run a checkpoint on count 0 (ie at the start).

    def step(self, quiet=True):
        """
        A single iteration of the manager's main loop.  Can be inter-mixed with add().
        """
        # the logic here is a little opaque because we need to mix layers of abstraction to
        # get the fine control we want for load balancing - we want to give jobs to whichever
        # source has least threads running so that the total number of workers (and so total
        # download bandwidth) is evenly spread across all servers.
        # a consequence of that is that there's no "step()" for lower levels.  this is also
        # partly because all the work is done in a separate worker process.  instead, most of
        # the lower level logic is done in clean_sources() which checks and updates the sources
        # appropriately (and, for single shot downloads, where quiet=False, raises an exception
        # on error).
        if not self._config_path:
            raise Exception('DownloadManager was created only to display data (no config_path)')
        # push any results from completed workers back to the sources
        self._workers.check()
        # and then update the state of the sources
        self._clean_sources(quiet=quiet)
        # with that done, update the stats for teh web display
        self._update_stats()
        # before trying to find a suitable candidates for more work...
        while self._workers.has_space() and self._has_data():
            # the order of sources is sorted here so that we round-robin consistently
            # (starting where we left off last time with self._index)
            sources = list(map(lambda name: self._source(name), sorted(self._sources.keys())))
            while True:
                # consider sources in turn
                source = self._next_source(sources)
                # are they deserving of an extra worker?
                if self._has_least_workers(source):
                    break
            # todo - why is this check needed?  shouldn't this always succeed?
            if source.has_chunks():
                source.new_worker(self._workers, self._config_path, self._rover_cmd)
                self._n_downloads += 1
            # todo - does this do anything useful without a workers.check()?
            self._clean_sources(quiet=quiet)

    def download(self):
        """
        Run to completion (for a single shot, after add()).
        """
        if len(self._sources) != 1:
            raise Exception('download() logic intended for single source (retrieve)')
        source = next(iter(self._sources.values()))
        try:
            while self._sources and not source.is_complete():
                self.step(quiet=False)
                sleep(0.1)
        finally:
            # not needed in normal use, as no workers when no sources, but useful on error
            self._workers.wait_for_all()

        return self._n_downloads

    # stats for web display

    def _create_stats_table(self):
        self.execute('''CREATE TABLE IF NOT EXISTS rover_download_stats (
                          submission text not null,
                          initial_stations int not null,
                          remaining_stations int not null,
                          initial_time float not null,
                          remaining_time float not null,
                          n_retries int not null,
                          download_retries int not null
                        )''')

    def _update_stats(self):
        with self._db:  # single transaction
            self._db.cursor().execute('BEGIN')
            self._db.execute('DELETE FROM rover_download_stats', tuple())
            for source in self._sources.values():
                progress = source.stats()
                self._db.execute('''INSERT INTO rover_download_stats
                                    (submission, initial_stations, remaining_stations, initial_time, remaining_time,
                                     n_retries, download_retries)
                                    VALUES (?, ?, ?, ?, ?, ?, ?)''',
                                 (source.name,
                                  progress.stations[1], progress.stations[1] - progress.stations[0],
                                  progress.seconds[1], max(0, int(progress.seconds[1] - progress.seconds[0])),
                                  source.n_retries, source.download_retries))

    def _start_web(self):
        if windows():
            cmd = 'pythonw -m rover %s -f %s %s 0' % (WEB, self._config_path, mm(VERBOSITY))
        else:
            cmd = '%s %s -f %s %s 0' % (self._rover_cmd, WEB, self._config_path, mm(VERBOSITY))
        run(cmd, self._log, uncouple=True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from builtins import range
from builtins import map
from future import standard_library
standard_library.install_aliases()
from random import randint, seed
from sys import version_info

if version_info[0] >= 3:
    from tempfile import TemporaryDirectory
else:
    from backports.tempfile import TemporaryDirectory

from rover.logs import init_log
from rover.coverage import Coverage
from rover.utils import format_epoch, parse_epoch

from .test_utils import WindowsTemp


def loop_date(d1, known=None):
    e = parse_epoch(d1)
    if known is not None:
        assert e == known, (e, known)
    d2 = format_epoch(e)
    assert d1 == d2, (d1, d2)


def test_epoch():
    loop_date('1970-01-01T00:00:00.000000', known=0)
    loop_date('2018-05-19T01:23:45.123456')


# rather than try to work out what tests are what in my head, the test
# framework itself prints an image of the test and the result.  hopefully
# the two can be checked by eye and then the result copied into the
# expected value.


def build_coverage(log, width, gap, offset):
    total = width + gap
    left = (offset % total) - width
    indices = []
    while left < 10:
        right = left + width
        indices.append((max(left, 0), min(right, 10)))
        left += total
    return indices


def coverage_to_str(coverage):
    return '(' + ','.join(map(lambda be: '(%g,%g)' % be, coverage.timespans)) + ')'


def indices_to_coverage(log, tolerance, increment, indices):
    coverage = Coverage(log, tolerance, increment, 'N.S.L.C')
    for (begin, end) in indices:
        if begin != end:
            coverage.add_epochs(begin, end, 1)
    return coverage


def include(indices, hour):
    for (begin, end) in indices:
        if begin <= hour <= end:
            return True
    return False


def print_labels():
    print('%9s ' % '', end='')
    for i in range(10):
        print('%d    ' % i, end='')
    print(10)


def print_coverage(title, coverage):
    print('%9s ' % title, end='')
    for i in range(0, 11):
        if i:
            if include(coverage.timespans, i-1) and include(coverage.timespans, i):
                print('====', end='')
            else:
                print('    ', end='')
        if include(coverage.timespans, i):
            print('#', end='')
        else:
            print('.', end='')
    print()


def run_explicit(log, tolerance, increment, index, avail, expected):
    """
    A bad test looks like:

          0    1    2    3    4    5    6    7    8    9    10
    index .    #====#    .    #====#    .    #====#    .    .
    avail .    .    #====#    .    #====#    .    #====#    .
  missing .    .    .    .    .    .    #    .    .    #    .

    where, for example, avail has data at 3 but it's not in missing
    """
    index = indices_to_coverage(log, tolerance, increment, index)
    avail = indices_to_coverage(log, tolerance, increment, avail)
    missing = avail.subtract(index)
    expected = indices_to_coverage(log, tolerance, increment, expected)
    if True or expected != missing:
        print()
        print_labels()
        print_coverage('index', index)
        print_coverage('avail', avail)
        print_coverage('missing', missing)
    assert expected == missing, coverage_to_str(missing)


def run(log, tolerance, increment,
        width_index, gap_index, offset_index,
        width_avail, gap_avail, offset_avail,
        expected):
    run_explicit(log, tolerance, increment,
                 build_coverage(log, width_index, gap_index, offset_index),
                 build_coverage(log, width_avail, gap_avail, offset_avail),
                 expected)


def test_coverage():
    with WindowsTemp(TemporaryDirectory) as dir:
        log = init_log(dir, '10M', 1, 5, 4, 'coverage', False, 1)[0]
        run(log, 0.5, 0.5, 0, 10, 0, 3, 2, 0, ((2,5),(7,10)))
        run(log, 0.5, 0.5, 2, 1, -1, 3, 2, 0, ())
        run(log, 0.5, 0.5, 2, 1, -1, 3, 2, 1, ())
        run(log, 0.5, 0.5, 2, 1, -1, 3, 2, 2, ())
        run(log, 0.5, 0.5, 2, 1, -1, 1, 1, 0, ())
        run(log, 0.5, 0.5, 2, 1, 0, 1, 1, 0, ((9.5,10),))
        run(log, 0.5, 0.5, 2, 1, 1, 1, 1, 0, ())
        run(log, 0.5, 0.5, 1, 2, 0, 1, 2, 0, ())
        run(log, 0.5, 0.5, 1, 2, 1, 1, 2, 0, ((2,2.5),(5,5.5),(8,8.5)))
        run(log, 0.5, 0.5, 1, 2, 2, 1, 2, 0, ((2.5,3),(5.5,6),(8.5,9)))
        run(log, 0.5, 0.5, 1, 2, 3, 1, 2, 0, ())
        run(log, 0.5, 0.5, 2, 2, 0, 2, 2, 0, ())
        run(log, 0.5, 0.5, 2, 2, 0, 2, 2, 1, ((0,1),(4.5,5),(8.5,9)))
        run(log, 0.5, 0.5, 2, 2, 0, 2, 2, 2, ((0,1.5),(4.5,5.5),(8.5,10)))
        run(log, 0.5, 0.5, 1, 1, 0, 3, 2, 0, ())
        run(log, 0.5, 0.5, 1, 1, 0, 3, 2, 1, ((0,0.5),))

# with tolerance 1.5 these overlap
# (this was a bug with rover retrieve IU_ANMO_3?_* 2016-01-01T20:00:00 2016-01-02T04:00:00)
#         0    1    2    3    4    5    6
#   index |    |====|====|====|    |    |
#   avail |    |    |    |====|====|    |
# missing |    |    |    |    |    |    |
        run_explicit(log, 1.5, 0.5, [(1,4)], [(3,5)], [])
# more of same (in retrospect these seem weird - we shouldn't be using a tolerance of 1.5!)
        seed(42)
        for i in range(100):
            # available of length at least 2 (so > tolerance)
            avail_start = randint(0, 4)
            avail_end = randint(avail_start + 2, 6)
            # index must start before one after start of avail (so before, within tolerance)
            index_start = randint(0, min(4, avail_start+1))
            # and must end with a width of at least 2, and at least one before end (so after, within tolerance)
            index_end = randint(max(avail_end-1, index_start+2), 6)
            run_explicit(log, 1.5, 0.5, [(index_start, index_end)], [(avail_start, avail_end)], [])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library
standard_library.install_aliases()
from sys import version_info
from os.path import join

from rover.args import DATADIR

if version_info[0] >= 3:
    from tempfile import TemporaryDirectory
else:
    from backports.tempfile import TemporaryDirectory

from rover.ingest import Ingester
from .test_utils import find_root, assert_files, TestConfig, WindowsTemp


def test_ingester():
    with WindowsTemp(TemporaryDirectory) as dir:
        root = find_root()
        config = TestConfig(dir)
        ingester = Ingester(config)
        ingester.run((join(root, 'tests', 'data', 'IU.ANMO.00-2010-02-27T06-30-00.000-2010-02-27T10-30-00.000.mseed'),))
        data_dir = config.arg(DATADIR)
        assert_files(data_dir, 'IU')
        assert_files(join(data_dir, 'IU'), '2010')
        assert_files(join(data_dir, 'IU', '2010'), '058')
        assert_files(join(data_dir, 'IU', '2010', '058'), 'ANMO.IU.2010.058')



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from builtins import open
from builtins import str
from future import standard_library
standard_library.install_aliases()
from builtins import object
from os.path import join, dirname
from sys import version_info

if version_info[0] >= 3:
    from tempfile import TemporaryDirectory
else:
    from backports.tempfile import TemporaryDirectory

from rover.config import BaseConfig, RepoInitializer
from rover.args import Arguments, TEMPDIR, DATADIR, FILE, INIT_REPOSITORY, DEFAULT_FILE
from rover.utils import canonify, windows

from .test_utils import WindowsTemp


class DummyLog(object):

    def debug(self, *args): pass
    def info(self, *args): pass
    def default(self, *args): pass
    def warn(self, *args): pass
    def error(self, *args): pass
    def critical(self, *args): pass


def test_write_config():
    with WindowsTemp(TemporaryDirectory) as dir:
        argparse = Arguments()
        args, config_path = argparse.parse_args([INIT_REPOSITORY])
        config = BaseConfig(DummyLog(), None, args, None, dir)
        RepoInitializer(config).run([dir])
        path = config.file(FILE)
        dirpath = dirname(path)
        with open(path, 'r') as input:
            contents = input.read()
            assert contents == \
'''# the data directory - data, timeseries.sqlite
data-dir='''+dirpath+'''/data
# maximum number of attempts to download data
download-retries=3
# number of download instances to run
download-workers=5
# output data format. Choose from "mseed" (miniSEED) or "asdf" (ASDF)
output-format=mseed
# name of ASDF file when ASDF output is specified
asdf-filename=asdf.h5
# station service url
station-url=http://service.iris.edu/fdsnws/station/1/query
# availability service url
availability-url=http://service.iris.edu/irisws/availability/1/query
# dataselect service url
dataselect-url=http://service.iris.edu/fdsnws/dataselect/1/query
# temporary storage for downloads
temp-dir='''+dirpath+'''/tmp
# directory for logs
log-dir='''+dirpath+'''/logs
# log verbosity (0-6)
log-verbosity=4
# console verbosity (0-6)
verbosity=4
# auto-start the download progress web server?
web=True
# port for HTTP server
http-port=8000
# address for completion status
email=
# address of SMTP server
smtp-address=localhost
# port for SMTP server
smtp-port=25
''', contents

def test_multiple_flags():
    argparse = Arguments()
    args, config_path = argparse.parse_args(['--index', '--no-index'])
    assert not args.index


def test_CONFIGDIR_start():
    with WindowsTemp(TemporaryDirectory) as dir:
        config = join(dir, DEFAULT_FILE)
        with open(config, 'w') as output:
            output.write('temp-dir=${CONFIGDIR}/foo\n')
            output.write('data-dir=$${CONFIGDIR}/foo\n')
        argparse = Arguments()
        args, config_path = argparse.parse_args(['-f', config])
        config = BaseConfig(None, None, args, None, dirname(config_path))
        print(config.dir(TEMPDIR))
        print(config.dir(DATADIR))
        print(canonify(dir + canonify(dir) + '/foo'))
        assert config.dir(TEMPDIR)
        assert config.dir(TEMPDIR) == canonify(dir + '/foo'), config.dir(TEMPDIR)
        assert config.dir(DATADIR)
        assert config.dir(DATADIR) == canonify(dir + canonify(dir) + '/foo'), config.dir(DATADIR)


def test_CONFIGDIR_middle():
    if not windows():  # gets confused over messed up path
        with WindowsTemp(TemporaryDirectory) as dir:
            dir = canonify(dir)
            config = join(dir, '.rover')
            with open(config, 'w') as output:
                output.write('temp-dir=xx${CONFIGDIR}/foo\n')
                output.write('data-dir=xx$${CONFIGDIR}/foo\n')
            argparse = Arguments()
            args, config_path = argparse.parse_args(['-f', config])
            config = BaseConfig(None, None, args, None, dirname(config_path))
            assert config.dir(TEMPDIR)
            assert config.dir(TEMPDIR) == canonify(join(dir, 'xx' + dir + '/foo')), config.dir(TEMPDIR)
            assert config.dir(DATADIR)
            assert config.dir(DATADIR) == canonify(dir + '/xx' + canonify(dir) + '/foo'), config.dir(DATADIR)


def test_CONFIGDIR_bad():
    with WindowsTemp(TemporaryDirectory) as dir:
        config = join(dir, '.rover')
        with open(config, 'w') as output:
            output.write('temp-dir=${FOO}\n')
        argparse = Arguments()
        args, config_path = argparse.parse_args(['-f', config])
        try:
            config = BaseConfig(None, None, args, None, dirname(config_path))
            config.arg(TEMPDIR)
        except Exception as e:
            assert 'does not exist' in str(e), str(e)
        else:
            assert False, 'Expected exception'



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import open
from future import standard_library
standard_library.install_aliases()
from io import StringIO
from os.path import join
from re import sub
from sys import version_info

if version_info[0] >= 3:
    from tempfile import TemporaryDirectory
else:
    from backports.tempfile import TemporaryDirectory

from rover.logs import init_log
from rover.args import DEFAULT_LOGVERBOSITY, DEFAULT_VERBOSITY

from .test_utils import WindowsTemp


def log_all(log):
    log.debug('debug')
    log.info('info')
    log.default('default')
    log.warn('warn')
    log.error('error')
    log.critical('critical')

def remove_timestamp(text):
    return sub(r'\d{4}\-\d{2}\-\d{2} \d{2}:\d{2}:\d{2},\d{3}', '[timestamp]', text)


def do_all_levels(log_level, log_expected, stderr_level, stderr_expected):

    with WindowsTemp(TemporaryDirectory) as dir:

        logdir = join(dir, '.rover')
        stderr = StringIO()
        log = init_log(logdir, '7M', 1, log_level, stderr_level, 'rover', False, 0, stderr=stderr)[0]
        log_all(log)
        stderr.seek(0)
        stderr_contents = stderr.read()
        assert stderr_contents == stderr_expected, stderr_contents

        with open(join(logdir, 'rover.log'), 'r') as output:
            log_contents = remove_timestamp(output.read())
            log_expected = remove_timestamp(log_expected)
            assert log_contents == log_expected, log_contents


def test_default_levels():
    do_all_levels(DEFAULT_LOGVERBOSITY,
'''DEFAULT  [timestamp]: default
WARNING  [timestamp]: warn
ERROR    [timestamp]: error
CRITICAL [timestamp]: critical
''',
                  DEFAULT_VERBOSITY,
'''rover  DEFAULT: default
rover  WARNING: warn
rover    ERROR: error
rover CRITICAL: critical
''')

def test_all_levels():
    do_all_levels(6,
'''DEBUG    [timestamp]: debug
INFO     [timestamp]: info
DEFAULT  [timestamp]: default
WARNING  [timestamp]: warn
ERROR    [timestamp]: error
CRITICAL [timestamp]: critical
''',
                  6,
'''rover    DEBUG: debug
rover     INFO: info
rover  DEFAULT: default
rover  WARNING: warn
rover    ERROR: error
rover CRITICAL: critical
''')

def test_critical_only():
    do_all_levels(1, '''CRITICAL [timestamp]: critical\n''', 1, '''rover CRITICAL: critical\n''')

def test_silent():
    do_all_levels(0, '''''', 0, '''''')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library
standard_library.install_aliases()
from sys import version_info
from os.path import join

from rover.args import DATADIR

if version_info[0] >= 3:
    from tempfile import TemporaryDirectory
else:
    from backports.tempfile import TemporaryDirectory

from rover.download import Downloader
from .test_utils import assert_files, TestConfig, WindowsTemp


def test_download():
    with WindowsTemp(TemporaryDirectory) as dir:
        config = TestConfig(dir)
        downloader = Downloader(config)
        downloader.run(['http://service.iris.edu/fdsnws/dataselect/1/query?net=IU&sta=ANMO&loc=00&cha=BHZ&start=2010-02-27T06:30:00.000&end=2010-02-27T10:30:00.000'])
        data_dir = config.arg(DATADIR)
        assert_files(join(data_dir, 'IU'), '2010')
        assert_files(join(data_dir, 'IU', '2010'), '058')
        assert_files(join(data_dir, 'IU', '2010', '058'), 'ANMO.IU.2010.058')
        n = config.db.cursor().execute('select count(*) from tsindex').fetchone()[0]
        assert n == 1, n



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from builtins import int
from future import standard_library
standard_library.install_aliases()
import sys
from re import match, sub
from sqlite3 import OperationalError

from .config import timeseries_db
from .args import MSEEDINDEXCMD, LEAP, LEAPEXPIRE, LEAPFILE, LEAPURL, DEV, VERBOSITY, MSEEDINDEXWORKERS, HTTPTIMEOUT, \
    HTTPRETRIES, FORCECMD, TIMESPANINC
from .args import TIMESPANTOL
from .coverage import MultipleSNCLBuilder
from .help import HelpFormatter
from .scan import ModifiedScanner, DirectoryScanner
from .sqlite import SqliteSupport
from .utils import format_epoch, windows, tidy_timestamp
from .utils import check_leap, check_cmd, STATION, NETWORK, CHANNEL, LOCATION
from .workers import Workers


"""
Commands related to the index:

The 'rover index' command - call mseeedindex to update the tsindex table.
The 'rover list-index' command - displays entries from the tsindex table.
"""


class Indexer(ModifiedScanner, DirectoryScanner):
    """
### Index

    rover index [--all]

    rover index (file|dir)+

Indexes files, adds or changes entries in the tsindex table stored in the
miniSEED database.

When no argument is given, all modified files in the repository are processed.
The `--all` flag forces all files to be processed. If a path argument
is provided, all files contained in the directory are processed, along with
the contents of sub-directories, unless `--no-recurse` is specified.

##### Significant Options

@all
@data-dir
@mseedindex-cmd
@mseedindex-workers
@leap
@leap-expire
@leap-file
@leap-url
@verbosity
@log-dir
@log-verbosity

##### Examples

    rover index --all

will index the entire repository.

"""

# Most of the work is done in the scanner superclasses which find the files
# to modify, and in the worker that runs mseedindex.

    def __init__(self, config):
        ModifiedScanner.__init__(self, config)
        DirectoryScanner.__init__(self, config)
        self._mseed_cmd = check_cmd(config, MSEEDINDEXCMD, 'mseedindex')
        self._timeseries_db = timeseries_db(config)
        self._leap_file = check_leap(config.arg(LEAP), config.arg(LEAPEXPIRE), config.arg(LEAPFILE),
                                     config.arg(LEAPURL), config.arg(HTTPTIMEOUT), config.arg(HTTPRETRIES), config.log)
        self._verbose = config.arg(DEV) and config.arg(VERBOSITY) == 5
        self._workers = Workers(config, config.arg(MSEEDINDEXWORKERS))

    def run(self, args):
        """
        Find the files to process (see superclasses).
        """
        if not args:
            self._log.info('Indexing all changed files')
            self.scan_data_dir()
        else:
            self.scan_dirs_and_files(args)

    def process(self, path):
        """
        Run mseedindex asynchronously in a worker.
        """
        self._log.info('Indexing %s' % path)
        if windows():
            self._workers.execute('set LIBMSEED_LEAPSECOND_FILE=%s && %s %s -sqlite %s %s'
                                  % (self._leap_file, self._mseed_cmd, '-v -v' if self._verbose  else '',
                                     self._timeseries_db, path))
        else:
            self._workers.execute('LIBMSEED_LEAPSECOND_FILE=%s %s %s -sqlite %s %s'
                                  % (self._leap_file, self._mseed_cmd, '-v -v' if self._verbose  else '',
                                     self._timeseries_db, path))

    def done(self):
        self._workers.wait_for_all()


START = 'start'
END = 'end'
COUNT = 'count'
JOIN = 'join'
JOIN_QSR = 'join-qsr'
QUALITY = 'quality'
SAMPLERATE = 'samplerate'


class IndexLister(SqliteSupport, HelpFormatter):
    """
### List Index

    rover list-index [net=...|sta=...|loc=...|cha=..|qua=...|samp=...]* [start=...] [end=...] \\
    [count|join|join-qsr]

    rover list-index [N_S_L_C_Q]* [start=...] [end=...] \\
    [count|join|join-qsr]

List an index of entries for a ROVER repository, defined by the the data-dir
configuration options, that match given constraints. For more information,
run "rover list-index" with no arguments.

Flag options used to change the output format are optional arguments.
Flags are mutually exclusive and take no value:

  count - only the number of matches will be shown

  join - continguous time ranges will be joined

  join-qsr - the maximal timespan across all quality and samplerates is shown
  (as used by retrieve)

##### Significant Options

@timespan-tol
@data-dir
@verbosity
@log-dir
@log-verbosity

#### Examples

    rover list-index IU_ANMO_00_BH? count

will display the number of entries for all time, and any quality or smaplerate.

    rover list-index net=* start=2001-01-01

will list all entries in the index after the year 2000.

"""

    def __init__(self, config):
        SqliteSupport.__init__(self, config)
        HelpFormatter.__init__(self, False)
        self._timespan_inc = config.arg(TIMESPANINC)
        self._timespan_tol = config.arg(TIMESPANTOL)
        self._timeseries_db = timeseries_db(config)
        self._multiple_constraints = {STATION: [],
                                      NETWORK: [],
                                      CHANNEL: [],
                                      LOCATION: [],
                                      QUALITY: [],
                                      SAMPLERATE: []}
        self._single_constraints = {START: None,
                                    END: None}
        self._flags = {COUNT: False, JOIN: False, JOIN_QSR: False}

    def _display_help(self):
        self.print_help('''
The list-index command prints entries from the index that match
the query parameters.  Parameters generally have the form
name=value (no spaces).

The following parameters take '*' and '?' as wildcards, can be
repeated for multiple matches (combined with 'OR"), and the name
only has to match unambiguously (so cha=HHZ is OK):

  station, network, channel, location, quality, samplerate

The short form N_S_L_C_Q can also be used.

The following parameters can be given only once, must be of
the form YYYY-MM-DDTHH:MM:SS.SSSSSS (may be truncated on the
right), and define a range of times over which the block must
appear (at least partially) to be included:

  start, end

The following parameters are simple flags that change the
output format.  They are mutually exclusive and take no
value:

  count - only the number of matches will be shown

  join - contiguous time ranges will be joined

  join-qsr - the maximal timespan across all quality and
  samplerates is shown (as used by retrieve)

Examples:

    rover list-index IU_ANMO_00_BH? count

will display the number of entries for all time, any
quality or smaplerate.

    rover list-index net=* start=2001-01-01

will list all entries in the index after the year 2000.

Note that console logging is to stderr, while results are
printed to stdout.

''')

    def run(self, args, stdout=None):
        if not stdout: stdout = sys.stdout
        if not args:
            self._display_help()
        else:
            self._check_database()
            self._parse_args(args)
            sql, params = self._build_query()
            if self._flags[COUNT]:
                self._count(sql, params, stdout=stdout)
            else:
                self._rows(sql, params, stdout=stdout)

    def _check_database(self):
        try:
            self.execute('SELECT count(*) FROM tsindex')
        except OperationalError:
            raise Exception('''Cannot access the index table in the database (%s).  Bad configuration or no data indexed?''' % self._timeseries_db)

    def _parse_args(self, args):
        for arg in args:
            if arg in self._flags:
                self._assert_unset_flags(arg)
                self._flags[arg] = True
            else:
                parts = arg.split('=')
                if len(parts) == 1:
                    self._set_snclq(arg)
                elif len(parts) != 2:
                    raise Exception('Cannot parse "%s" (not of form name=value)' % arg)
                else:
                    name, value = parts
                    if name in (START, END):
                        self._set_time_limit(name, value)
                    else:
                        self._set_name_value(name, value)
        if self._flags[JOIN_QSR] and (
                self._multiple_constraints[SAMPLERATE] or self._multiple_constraints[QUALITY]):
            raise Exception('Cannot specify %s / %s AND %s' % (QUALITY, SAMPLERATE, JOIN_QSR))

    def _assert_unset_flags(self, name1):
        for name2 in self._flags:
            if self._flags[name2]:
                raise Exception('Cannot specify multiple keys (%s, %s)' % (name1, name2))

    def _set_snclq(self, snclq):
        components = snclq.split('_')
        if len(components) < 2 or len(components) > 5:
            raise Exception('Cannot parse %s (expect 2 to 5 values separated by "_" for N_S_L_C_Q)' % snclq)
        self._set_name_value(NETWORK, components[0])
        self._set_name_value(STATION, components[1])
        if len(components) > 2: self._set_name_value(LOCATION, components[2])
        if len(components) > 3: self._set_name_value(CHANNEL, components[3])
        if len(components) > 4: self._set_name_value(QUALITY, components[4])

    def _set_time_limit(self, name, value):
        if self._single_constraints[name]:
            raise Exception('Multiple values for %s' % name)
        value = tidy_timestamp(self._log, value)
        self._single_constraints[name] = value
        if self._single_constraints[START] and self._single_constraints[END] \
                and self._single_constraints[START] > self._single_constraints[END]:
            raise Exception('start time (%s) must be after end time (%s)'
                            % (self._single_constraints[START], self._single_constraints[END]))

    def _set_name_value(self, name, value):
        if not match('^[\w\*\?]*$', value):
            raise Exception('Illegal characters in "%s"' % value)
        found = None
        for key in self._multiple_constraints.keys():
            if key.startswith(name):
                if found:
                    raise Exception('Ambiguous parameter: %s' % name)
                else:
                    found = key
        if not found:
            raise Exception('Did not recognise "%s"' % name)
        self._multiple_constraints[found].append(value)

    def _build_query(self):
        sql, params = 'select ', []
        if self._flags[COUNT]:
            sql += 'count(*) '
        else:
            sql += ('''network, station, location, channel,
                       coalesce(timespans,'<' || starttime || ' ' || endtime || '>'), samplerate ''')
            if not self._flags[JOIN_QSR]:
                sql += ', quality '
        sql += 'from tsindex '
        constrained = False

        def conjunction(sql, constrained):
            if constrained:
                sql += 'and '
            else:
                sql += 'where '
                constrained = True
            return sql, constrained

        for name in self._multiple_constraints:
            repeated = False
            for value in self._multiple_constraints[name]:
                if not repeated:
                    sql, constrained = conjunction(sql, constrained)
                    sql += '( '
                    repeated = True
                else:
                    sql += 'or '
                sql += '%s like ? ' % name
                params.append(self._wildchars(value))
            if repeated:
                sql += ') '
        if self._single_constraints[START]:
            sql, constrained = conjunction(sql, constrained)
            sql += 'endtime > ?'
            params.append(self._single_constraints[START])
        if self._single_constraints[END]:
            sql, constrained = conjunction(sql, constrained)
            sql += 'starttime < ?'
            params.append(self._single_constraints[END])
        return sql, tuple(params)

    @staticmethod
    def _wildchars(value):
        return sub(r'\?', '_', sub(r'\*', '%', value))

    def _count(self, sql, params, stdout):
        # force to int here to avoid issues with strings on python 2
        print(int(self.fetchsingle(sql, params)), file=stdout)

    def _rows(self, sql, params, stdout):
        self._log.debug('%s %s' % (sql, params))
        builder = MultipleSNCLBuilder(self._log, self._timespan_tol, self._timespan_inc,
                                      self._flags[JOIN] or self._flags[JOIN_QSR])

        def callback(row):
            if self._flags[JOIN_QSR]:
                n, s, l, c, ts, r = row
                builder.add_timespans('%s_%s_%s_%s' % (n, s, l, c), ts, r)
            else:
                n, s, l, c, ts, r, q = row
                builder.add_timespans('%s_%s_%s_%s_%s (%g Hz)' % (n, s, l, c, q, r), ts, r)

        self.foreachrow(sql, params, callback)
        print()
        for coverage in builder.coverages():
            print('  %s' % coverage.sncl)
            for ts in coverage.timespans:
                print('    %s - %s' % (format_epoch(ts[0]), format_epoch(ts[1])))
            print()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import open
from builtins import str
from future import standard_library
standard_library.install_aliases()
from datetime import datetime
from os import getpid
from os.path import exists, join
from re import match
from shutil import copyfile

from .args import MSEEDINDEXCMD, LEAP, LEAPEXPIRE, LEAPFILE, LEAPURL, \
    DATADIR, INDEX, HTTPTIMEOUT, HTTPRETRIES, OUTPUT_FORMAT
from .index import Indexer
from .lock import DatabaseBasedLockFactory, MSEED
from .scan import DirectoryScanner
from .sqlite import SqliteSupport, SqliteContext
from .utils import run, check_cmd, check_leap, create_parents, safe_unlink, \
    windows, atomic_move, hash

"""
The 'rover ingest' command - copy downloaded data into the repository (and then call index).
"""


# this is the table name when run directly from the command line.
# when run as a worker from (multiple) retriever(s) a table is supplied.
TMPFILE = 'rover_tmp_ingest'


class Ingester(SqliteSupport, DirectoryScanner):
    """
### Ingest

    rover ingest file

Adds contents from a miniSEED formatted file to ROVER's local repository and
indexes the new data.

##### Significant Options

@mseedindex-cmd
@data-dir
@index
@leap
@leap-expire
@leap-file
@leap-url
@verbosity
@log-dir
@log-verbosity

Options used to configure the sub-command index are also applicable -
see Index help for more details.

##### Examples

    rover ingest /tmp/IU.ANMO.00.*.mseed

will add all the data in the given file to the repository.

"""

# The simplest possible ingester:
# * Uses mseedindex to parse the file.
# * For each section, appends to any existing file using byte offsets
# * Refuses to handle blocks that cross day boundaries
# * Does not check for overlap, differences in sample rate, etc.

    def __init__(self, config):
        SqliteSupport.__init__(self, config)
        DirectoryScanner.__init__(self, config)
        self._mseed_cmd = check_cmd(config, MSEEDINDEXCMD, 'mseedindex')
        self._leap_file = check_leap(config.arg(LEAP), config.arg(LEAPEXPIRE), config.arg(LEAPFILE),
                                     config.arg(LEAPURL), config.arg(HTTPTIMEOUT), config.arg(HTTPRETRIES), config.log)
        self._db_path = None
        self._data_dir = config.dir(DATADIR)
        self._index = config.arg(INDEX)
        self._config = config
        self._log = config.log
        self._lock_factory = DatabaseBasedLockFactory(config, MSEED)

    def run(self, args, db_path=TMPFILE):
        """
        We only support explicit paths - modified file scanning makes no sense because
        the files are external, downloaded data.

        To avoid database contention we can run mseedindex into a unique database path.
        """
        self._db_path = db_path
        if not args:
            raise Exception('No paths provided')
        self.scan_dirs_and_files(args)

    def process(self, temp_file):
        """
        Run mseedindex and, move across the bytes, and then call follow-up tasks.
        """
        self._log.info('Indexing %s for ingest' % temp_file)
        if exists(self._db_path):
            self._log.warn('Temp file %s exists (deleting)' % self._db_path)
            safe_unlink(self._db_path)
        updated = set()
        try:
            if windows():
                run('set LIBMSEED_LEAPSECOND_FILE=%s && %s -sqlite %s %s'
                    % (self._leap_file, self._mseed_cmd, self._db_path, temp_file), self._log)
            else:
                run('LIBMSEED_LEAPSECOND_FILE=%s %s -sqlite %s %s'
                    % (self._leap_file, self._mseed_cmd, self._db_path, temp_file), self._log)
            with SqliteContext(self._db_path, self._log) as db:
                rows = db.fetchall('''SELECT network, station, starttime, endtime, byteoffset, bytes
                                  FROM tsindex ORDER BY byteoffset''')
                updated.update(self._copy_all_rows(temp_file, rows))
        finally:
            safe_unlink(self._db_path)
        if self._index:
            Indexer(self._config).run(updated)
            if self._config.arg(OUTPUT_FORMAT).upper() == "ASDF":
                from .asdf import ASDFHandler
                # output as ASDF format
                ASDFHandler(self._config).load_miniseed(updated)

    def _copy_all_rows(self, temp_file, rows):
        self._log.info('Ingesting %s' % temp_file)
        updated = set()
        self.updated_files_last_run = []
        with open(temp_file, 'rb') as input_file:
            offset = 0
            for row in rows:
                offset, dest = self._copy_single_row(offset, input_file,
                                                     temp_file, *row)
                updated.add(dest)
        return updated

    def _copy_single_row(self, offset, input_buffer, temp_file, network, station, starttime, endtime, byteoffset, raw_bytes):
        self._assert_single_day(temp_file, starttime, endtime, "%s_%s" % (network, station))
        if offset < byteoffset:
            self._log.warn('Non-contiguous bytes in %s - skipping %d bytes' % (temp_file, byteoffset - offset))
            input_buffer.seek(byteoffset - offset, 1)
            offset = byteoffset
        elif offset > byteoffset:
            raise Exception('Overlapping blocks in %s, index is inconsistent regarding byte ranges)' % temp_file)
        data = input_buffer.read(raw_bytes)
        offset += raw_bytes
        dest = self._make_destination(network, station, starttime)
        self._log.debug('Appending %d bytes from %s at offset %d to %s' % (raw_bytes, temp_file, byteoffset, dest))
        self._append_data(data, dest)
        return offset, dest

    def _make_destination(self, network, station, starttime):
        date_string = match(r'\d{4}-\d{2}-\d{2}', starttime).group(0)
        time_data = datetime.strptime(date_string, '%Y-%m-%d').timetuple()
        year, day = time_data.tm_year, time_data.tm_yday
        return join(self._data_dir, network, str(year), '%03d' % day, '%s.%s.%04d.%03d' % (station, network, year, day))

    def _append_data(self, data, mseed_file):
        # here we are locking for this process, so we can set the PID directly.
        # there is no possibility for deadlock because we are single threaded
        # and release on exit.
        with self._lock_factory.lock(mseed_file, pid=getpid()):
            # to avoid leaving broken files on unexpected exit, use a temp
            # file and then move into position (move should be atomic)
            tmp = mseed_file + '.tmp'
            if exists(tmp):
                self._log.warn('Cleaning %s' % tmp)
                safe_unlink(tmp)
            if not exists(mseed_file):
                create_parents(mseed_file)
                open(tmp, 'w').close()
            else:
                copyfile(mseed_file, tmp)
            with open(tmp, 'ba') as output:
                output.write(data)
            atomic_move(self._log, tmp, mseed_file)

    def _assert_single_day(self, temp_file, starttime, endtime, sid):
        # Comparing time strings, presumed format 'YYYY-MM-DDThh:mm:ss.ssssss'
        if starttime[:10] != endtime[:10]:
            raise Exception('File %s contains data from more than one day (%s-%s) for %s' % (temp_file, starttime, endtime, sid))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
import os
from .args import FORCE_METADATA_RELOAD, UserFeedback, \
    fail_early
from .config import asdf_container, timeseries_db
from .utils import safe_unlink
from .sqlite import SqliteContext
from .lock import DatabaseBasedLockFactory, ASDF

try:
    import pyasdf
    import obspy
except ImportError:
    raise Exception("Missing required 'pyasdf' python package for "
                    "'output-format=asdf.'") 

class ASDFHandler(UserFeedback):
    
    def __init__(self, config):
        UserFeedback.__init__(self, config)
        fail_early(config)
        self.asdf_path = asdf_container(config)
        self._config = config
        self._lock_factory = DatabaseBasedLockFactory(config, ASDF)
        
    def _get_asdf_dataset(self):
            try:
                ds = pyasdf.ASDFDataSet(self.asdf_path, mode="a")
                self._log.debug("Opened ASDF dataset '{}'."
                                .format(self.asdf_path))
            except Exception as e:
                raise Exception("Failed to create ASDF file at '{}': {}."
                                .format(self.asdf_path, e))
            return ds
        
    def load_miniseed(self, mseed_file_list):
        """
        Load miniSEED data into the ASDF archive and remove miniSEED file.
        """
        lock = self._lock_factory.lock(self.asdf_path, pid=os.getpid())
        lock.acquire()
        with SqliteContext(
                timeseries_db(self._config), self._log) as db:
            for mseed_file in mseed_file_list:
                # a context manager is used to ensure that the ASDF dataset
                # object is deleted so that other processes can use the file 
                with self._get_asdf_dataset() as ds:
                    st = obspy.read(mseed_file)
                    for trace in st:
                        # write timeseries to asdf file
                        self._log.default("Add '{}' to ASDF."
                                          .format(trace))
                        ds.append_waveforms(trace,
                                            tag="raw_recording")
                        # update miniSEED TSIndex records
                        # (format= “ASDF”, filename=<ASDF_FILENAME>)
                        db.execute("UPDATE tsindex "
                                   "SET filename='{0}', format='ASDF', "
                                   "byteoffset=null, hash=null "
                                   "WHERE filename='{1}'"
                                   .format(self.asdf_path, mseed_file))
                    # remove miniseed that was inserted into ASDF
                    safe_unlink(mseed_file)
        lock.release()
        
        
    def load_metadata(self, source):
        lock = self._lock_factory.lock(self.asdf_path, pid=os.getpid())
        lock.acquire()
        summary_rows = source.fetch_summary_rows()
        # group summary rows by ASDF waveform tag
        grouped_rows = {}
        for row in summary_rows:
            tag = "{}.{}".format(row.network, row.station)
            if not grouped_rows.get(tag):
                grouped_rows[tag] = [row]
            else:
                grouped_rows[tag].append(row)
    
        with self._get_asdf_dataset() as ds:
            # loop over grouped summary rows and compare against ASDF data
            for tag, rows in grouped_rows.items():
                if tag not in ds.waveforms:
                    # this should never happen
                    self._log.error("Skipping station '{0}'. Found station "
                                    "'{0}' in the tsindex_summary table but "
                                    "not in the ASDF container.".format(tag))
                    continue
                if "StationXML" in ds.waveforms[tag]:
                    # read StationXML inventory from ASDF
                    asdf_sta_inv = ds.waveforms[tag]["StationXML"]
                else:
                    # no StationXML was loaded for tag
                    asdf_sta_inv = None
                for row in rows:
                    # check if indexed channel exists in ASDF container
                    if self._config.arg(FORCE_METADATA_RELOAD) is True \
                            or asdf_sta_inv is None \
                            or not asdf_sta_inv.select(
                                network=row.network,  station=row.station,
                                location=row.location, channel=row.channel,
                                starttime=row.earliest, endtime=row.latest):
                        self._log.default("Load metadata for "
                                          "'{}.{}.{}.{} | {} - {}' "
                                          "into ASDF."
                                          .format(row.network, row.station,
                                                  row.location, row.channel,
                                                  row.earliest, row.latest))
                        # download StationXML from fdsnws station service
                        station_xml = source.download_stationxml(row.network,
                                                                 row.station,
                                                                 row.location,
                                                                 row.channel,
                                                                 row.earliest,
                                                                 row.latest)
                        inv = obspy.read_inventory(station_xml)
                        ds.add_stationxml(inv)
                        safe_unlink(station_xml)
                    else:
                        self._log.default("Skip ASDF metadata load for "
                                          "'{}.{}.{}.{} | {} - {}'. "
                                          "Metadata was already loaded."
                                          .format(row.network, row.station,
                                                  row.location, row.channel,
                                                  row.earliest, row.latest))
        lock.release()



from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from builtins import map
from builtins import next
from future import standard_library
standard_library.install_aliases()
from builtins import object
from .utils import PushBackIterator, format_epoch, parse_epoch

"""
Interface to the N_S_L_C / timespan data in tsindex - how much data do we have
for particular channels?
"""


class Coverage(object):
    """
    The data coverage (timespans) with the index for a given N_S_L_C.

    This includes the logic to merge timespans.  The N_S_L_C parameter is
    a free string (or arbitrary object) and so this class can be used
    for N_S_L_C, N_S_L_C + sampling rate, etc.
    """

    def __init__(self, log, frac_tolerance, frac_increment, sncl):
        self._log = log
        self._frac_tolerance = frac_tolerance
        self._frac_increment = frac_increment
        self.sncl = sncl
        self.timespans = []
        self.samplerate = None

    def add_samplerate(self, samplerate):
        if samplerate is not None:
            if self.samplerate is not None:
                self.samplerate = min(self.samplerate, samplerate)
            else:
                # a samplerate of zero seems to be used for logging or other
                # state-of-health channels, so use a large value in that case
                self.samplerate = samplerate if samplerate else 10000

    def add_epochs(self, start, end, samplerate=None):
        """
        Add a timespan to those that already exist, merging if necessary.

        IMPORTANT: This assumes that the timestamps are added ordered
        by increasing start time - see builders that guarantee that.
        """
        self.add_samplerate(samplerate)
        self.timespans.append((start, end))

    def join(self):
        self._log.debug('Joining overlapping timespans')
        if self:  # avoid looking at samplerate if no data
            joined, (tolerance, increment) = [], self.tolerances()
            for start, end in self.timespans:
                if not joined:
                    joined.append((start, end))
                else:
                    b, e = joined[-1]
                    if start < b:
                        raise Exception('Unsorted start times')
                    # do they overlap at all?
                    if self.samplerate == 0:
                        # Channels with 0 sample rate must always be merged.
                        self._log.debug('Joining channel with sample rate of zero.')
                        joined[-1]=(b, end)
                    elif abs(start - e) < 1.0 / self.samplerate + tolerance:
                        # if they do, and this extends previous, replace with maximal span
                        if end > e:
                            self._log.debug('Joining %d-%d and %d-%d' % (start, end, b, e))
                            joined[-1] = (b, end)
                    # no they don't overlap
                    else:
                        joined.append((start, end))
            self.timespans = joined

    def __str__(self):
        return '%s: %d timespans from %s to %s' % (
            self.sncl, len(self.timespans),
            format_epoch(self.timespans[0][0]) if self.timespans else '-',
            format_epoch(self.timespans[-1][1]) if self.timespans else '-'
        )

    def __eq__(self, other):
        try:
            return self.sncl == other.sncl and self.timespans == other.timespans
        except:
            return False

    def __bool__(self):
        return bool(self.timespans)

    def tolerances(self):
        if self.samplerate is None:
            raise Exception('Sample rate is not available.')
        if self.samplerate == 0:
            return 0.0, 0.0
        return self._frac_tolerance / self.samplerate, self._frac_increment / self.samplerate

    def subtract(self, other):
        """
        Calculate the timespans for which this instance has data, but the
        other instance does not.
        """
        if not self.sncl == other.sncl:
            raise Exception('Cannot subtract mismatched availabilities')
        if not other:  # subtracting zero (avoid checking samplerate)
            return self

        # minimal samplerate
        self.add_samplerate(other.samplerate)
        other.add_samplerate(self.samplerate)
        tolerance, increment = self.tolerances()
        # simplify the work by compressing both
        self.join()
        other.join()

        us, them = PushBackIterator(iter(self.timespans)), PushBackIterator(iter(other.timespans))
        difference = Coverage(self._log, self._frac_tolerance, self._frac_increment, self.sncl)
        difference.add_samplerate(self.samplerate)
        while True:
            try:
                us_start, us_end = next(us)
            except StopIteration:
                # we can return difference now, because there's only more subtracting to do
                return difference
            try:
                them_start, them_end = next(them)
            except StopIteration:
                # there's no more subtraction, so everything left goes into difference
                if us_end - us_start >= tolerance:
                    difference.add_epochs(us_start, us_end)
                for (us_start, us_end) in us:
                    if us_end - us_start >= tolerance:
                        difference.add_epochs(us_start, us_end)
                return difference

            # we start together
            if abs(us_start - them_start) < tolerance:
                # if we end together too, there's nothing to do
                if abs(us_end - them_end) < tolerance:
                    pass
                # if we end first, so are completely wiped out, while they live to
                # perhaps delete more
                elif us_end < them_end:
                    them.push((them_start, them_end))
                # but they end first.  so some of our timespan still lives to face
                # the next challenger.
                else:
                    # since this is a real difference we need to nudge forwards so
                    # that we avoid including the end point of them again.
                    us.push((them_end + increment, us_end))
            # we start before them (difference must be larger than tolerance - see above)
            elif us_start < them_start:
                # we also end before them, so we're home free into the difference and
                # they live to try kill our next timespan
                if us_end < them_start:
                    difference.add_epochs(us_start, us_end)
                    them.push((them_start, them_end))
                # we end after they start, so we overlap.  save the initial part in
                # the difference and push the rest back for further consideration.
                else:
                    # is (us_start, them_start - increment) worth adding?
                    if them_start - increment - us_start >= tolerance:
                        difference.add_epochs(us_start, them_start - increment)
                    if us_end - them_start > tolerance:
                        us.push((them_start, us_end))
                    them.push((them_start, them_end))
            # we start after them
            else:
                # if we also end before them, then we're deleted completely
                # while they continue to face our next timespan.
                # (this is tricky - we can end slightly after, if it's within tolerance)
                if us_end - them_end < tolerance:
                    them.push((them_start, them_end))
                # but we also end after them.  so some (perhaps all) of our timespan
                # remains to face their next timespan.
                else:
                    # again, we need to nudge forwards beyond them to avoid
                    # re-including the point
                    us.push((max(them_end + increment, us_start), us_end))


# builders are needed to buffer the data read from the database and sort it,
# this is because:
# (1) the add_epochs() method requires sorted data to correctly merge timespans
# (2) the timespans are squashed into strings that are not easy to use at the SQL
#     level (so we need to read and then sort in-memory).


class BaseBuilder(object):
    """
    Shared functionality for all coverage builders.
    """

    def __init__(self, log, frac_tolerance, frac_increment):
        self._log = log
        self._frac_tolerance = frac_tolerance
        self._frac_increment = frac_increment

    def _parse_timespans(self, timespans):
        if timespans is None:
            raise Exception('Unexpected NULL reading timespans')
        for pair in timespans.split(','):
            inner = pair[1:-1]
            if pair[0] == '[':
                start, end = map(float, inner.split(':'))
            elif pair[0] == '<':
                start, end = map(parse_epoch, inner.split(' '))
            else:
                raise Exception('Unexpected timespans format: "%s"' % pair)
            yield start, end


class SingleSNCLBuilder(BaseBuilder):
    """
    Sort the timestamp information before creating the coverage (for a single N_S_C_C)

    The mseedindex schema design makes it difficult to sort this information in SQL.
    """

    def __init__(self, log, frac_tolerance, frac_increment, sncl):
        super().__init__(log, frac_tolerance, frac_increment)
        self._sncl = sncl
        self._timespans = []

    def add_timespans(self, timespans, samplerate=None):
        for start, end in self._parse_timespans(timespans):
            self._timespans.append((start, end, samplerate))

    def coverage(self):
        coverage = Coverage(self._log, self._frac_tolerance, self._frac_increment, self._sncl)
        for start, end, samplerate in sorted(self._timespans):
            coverage.add_epochs(start, end, samplerate)
        return coverage


class MultipleSNCLBuilder(BaseBuilder):
    """
    Sort the timestamp information before creating the coverage (for multiple N_S_L_Cs).

    The mseedindex schema design makes it difficult to sort this information in SQL.
    """

    def __init__(self, log, frac_tolerance, frac_increment, join=True):
        super().__init__(log, frac_tolerance, frac_increment)
        self._join = join
        self._timespans = {}

    def add_timespans(self, sncl, timespans, samplerate=None):
        if sncl not in self._timespans:
            self._timespans[sncl] = []
        ts = self._timespans[sncl]
        for start, end in self._parse_timespans(timespans):
            ts.append((start, end, samplerate))

    def coverages(self):
        for sncl in sorted(self._timespans.keys()):
            ts = self._timespans[sncl]
            coverage = Coverage(self._log, self._frac_tolerance, self._frac_increment, sncl)
            for start, end, samplerate in sorted(ts):
                coverage.add_epochs(start, end, samplerate)
            if self._join:
                coverage.join()
            yield coverage
